"""An interface for creating and retrieving Matchlight projects."""
from __future__ import absolute_import

import datetime
try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc

import json

import six

import matchlight.error
import matchlight.utils


class Project(object):
    """A Matchlight Fingerprint Monitoring Project.

    Attributes:
        name (:obj:`str`): The project name.
        project_type (:obj:`str`): The project type.
        upload_token (:obj:`str`): The project upload token.
        last_date_modified (:obj:`int`): The Unix timestamp of the last
            modification.
        number_of_records (:obj:`int`): The number of total records in
            the project.
        number_of_unseen_alerts (:obj:`int`): The number of unread
            alerts.

    """

    def __init__(self, name, project_type, upload_token,
                 last_date_modified, number_of_records,
                 number_of_unseen_alerts):
        """Initializes a new project.

        Args:
            name (:obj:`str`): The project name.
            project_type (:obj:`str`): The project type.
            upload_token (:obj:`str`): The project upload token.
            last_date_modified (:obj:`int`): The Unix timestamp of the
                last modification.
            number_of_records (:obj:`int`): The number of total records
                in the project.
            number_of_unseen_alerts (:obj:`int`): The number of unread
                alerts.

        """
        self.name = name
        self.project_type = project_type
        self.upload_token = upload_token
        self.last_date_modified = last_date_modified
        self.number_of_records = number_of_records
        self.number_of_unseen_alerts = number_of_unseen_alerts

    @classmethod
    def from_mapping(cls, mapping):
        """Creates a new project instance from the given mapping."""
        return cls(
            name=mapping['project_name'],
            project_type=mapping['project_type'],
            upload_token=mapping['project_upload_token'],
            last_date_modified=mapping['last_date_modified'],
            number_of_records=mapping['number_of_records'],
            number_of_unseen_alerts=mapping['number_of_unseen_alerts'],
        )

    @property
    def last_modified(self):
        """:class:`datetime.datetime`: The last modified timestamp."""
        return datetime.datetime.fromtimestamp(self.last_date_modified)

    @property
    def details(self):
        """:obj:`dict`: Returns the project details as a mapping."""
        return {
            'name': self.name,
            'project_type': self.project_type,
            'upload_token': self.upload_token,
            'last_date_modified': self.last_date_modified,
            'number_of_records': self.number_of_records,
            'number_of_unseen_alerts': self.number_of_unseen_alerts,
        }

    def __repr__(self):  # pragma: no cover
        return '<Project(name="{}", project_type="{}")>'.format(
            self.name, self.project_type)


class ProjectMethods(collections_abc.Iterable):
    """Provides methods for interfacing with the feeds API.

    Examples:
        Get project from upload token::
            >>> ml.projects.get('3ef85448c-d244-431e-a207-cf8d37ae3bfe')
            <Project(name='Customer Database May 2016',
            project_type='pii')>

        Filter on project types::
            >>> ml.projects.filter(project_type='pii')
            [<Project(name='...', project_type='pii'),
            <Project(name='...', project_type='pii'),
            <Project(name='...', project_type='pii')]
            >>> ml.projects.filter()
            [<Project(name='...', project_type='pii'),
            <Project(name='...', project_type='document'),
            <Project(name='...', project_type='source_code')]

        Create a new project::
            >>> project = ml.projects.add(
            ... name='Super secret algorithm',
            ... project_type='source_code')
            >>> project
            <Project(name='Super Secret Algorithm', type='source_code')>

        Edit a project::
            >>> project = ml.projects.edit(project,
            ... 'Updated Super Secret Algorithm')
            >>> project
            <Project(name='Updated Super Secret Algorithm',
            type='source_code'>

        Delete a project::
            >>> ml.projects.delete(project)
            >>> ml.projects.get(project.upload_token)
            None

        Get project details::
            >>> executives = ml.projects.add("Executive List 2016", "pii")
            >>> executives.details
            {'last_date_modified': 1472671877,
            'number_of_records': 0,
            'number_of_unseen_alerts': 0,
            'name': 'Executive List 2016',
            'project_type': 'pii',
            'upload_token': 'a1c7140a-17e5-4016-8f0a-ef4aa87619ce'}

    """

    def __init__(self, ml_connection):  # noqa: D205,D400
        """Initializes a project interface with the given Matchlight
        connection.

        Args:
            ml_connection (:class:`~.Connection`): A Matchlight
                connection instance.

        """
        self.conn = ml_connection

    def all(self):
        """Returns all projects associated with the account."""
        return self.filter()

    def add(self, name, project_type):
        """Creates a new project or group.

        Arguments:
            name (:obj:`str`): The name of the project to be created.
            project_type (:obj:`str`): The type of project to be
                created.

        Returns:
            :class:`~.Project` : Created project with upload token.

        """
        data = json.dumps({'name': name, 'type': project_type})
        r = self.conn.request('/project/add', data=data)
        return Project.from_mapping(r.json()['data'])

    def delete(self, project):
        """Deletes a project and all associated records.

        Args:
            project (:class:`~.Project` or :obj:`str`): The project
                object or upload token to be deleted.

        """
        if isinstance(project, six.string_types):
            upload_token = project
        else:
            upload_token = project.upload_token
        self.conn.request('/project/{upload_token}/delete'.format(
            upload_token=upload_token), data='{}')

    def edit(self, project, updated_name):
        """Renames a project.

        Arguments:
            project (:class:`~.Project` or :obj:`str`): A project
                instance or upload token.
            updated_name (:obj:`str`): New project name.

        Returns:
            :class:`~.Project`: Updated project instance with new name.

            Note that this method mutates any project instances passed.

        """
        if not isinstance(project, Project):
            project = self.get(project)
        data = json.dumps({'name': updated_name})
        self.conn.request('/project/{}/edit'.format(
            project.upload_token), data=data)
        project.name = updated_name
        return project

    def filter(self, project_type=None):
        """Returns a list of projects associated with the account.

        Providing an optional **project_type** keyword argument will
        only return projects of the specified type: ``source_code``,
        ``document`` or ``pii``.

        Arguments:
            project_type (:obj:`str`, optional): The project type to
                filter on. If not provided or ``None``, returns all
                projects.

        Returns:
            list of :class:`~.Project`: List of filtered projects.

        """
        response = self.conn.request('/projects', params={
            'project_type': project_type})
        projects = []
        for payload in response.json().get('data', []):
            if project_type and payload['project_type'] != project_type:
                continue
            projects.append(Project.from_mapping(payload))
        return projects

    def get(self, upload_token):
        """Returns a project by the given upload token.

        Args:
            upload_token (:obj:`str`): The project upload token.

        Returns:
           :class: `~.Project`: A Matchlight project.

        """
        try:
            response = self.conn.request('/project/{}'.format(upload_token))
            return Project.from_mapping(response.json())
        # for consistency since records.get() returns None if not found.
        except matchlight.error.APIError as err:
            if err.args[0] == 404:
                return
            else:
                raise

    def __iter__(self):
        return (project for project in self.filter())



"""Unit tests Matchlight SDK connection objects."""
import pytest

import matchlight


def test_invalid_connection():
    """Verify SDK error thrown when initializing an invalid connection."""
    with pytest.raises(matchlight.SDKError):
        matchlight.Matchlight(access_key=None, secret_key=None)


def test_connection_with_environment(monkeypatch, access_key, secret_key):
    """Verify the connection uses access and secret keys from environment."""
    monkeypatch.setenv('MATCHLIGHT_ACCESS_KEY', access_key)
    monkeypatch.setenv('MATCHLIGHT_SECRET_KEY', secret_key)
    ml = matchlight.Matchlight()
    assert ml.conn.access_key == access_key
    assert ml.conn.secret_key == secret_key



"""Software development kit for the Terbium Labs Matchlight product."""
from __future__ import absolute_import

import pkg_resources

from .alert import Alert, AlertMethods
from .connection import Connection, MATCHLIGHT_API_URL_V2
from .error import (
    APIError,
    ConnectionError,
    InvalidCredentialsError,
    SDKError,
)
from .feed import Feed, FeedMethods
from .project import Project, ProjectMethods
from .record import Record, RecordMethods
from .search import SearchMethods


__all__ = (
    'Alert',
    'AlertMethods',
    'APIError',
    'Connection',
    'ConnectionError',
    'Feed',
    'FeedMethods',
    'InvalidCredentialsError',
    'MATCHLIGHT_API_URL_V2',
    'Matchlight',
    'Project',
    'ProjectMethods',
    'Record',
    'RecordMethods',
    'SDKError',
    'SearchMethods',
)


try:
    __version__ = pkg_resources.get_distribution('matchlightsdk').version
except:
    __version__ = 'unknown'


class Matchlight(object):
    """Top-level Matchlight API connection object.

    A top-level wrapper for all Matchlight products: including
    Retrospective Search, DataFeeds, and Fingerprint Monitoring
    (projects, and records).

    Attributes:
        alerts: :class:`~matchlight.alert.AlertMethods` object
            with access to alert methods for Matchlight Fingerprint
            Monitoring.
        projects: :class:`~matchlight.project.ProjectMethods` object
            with access to project methods for Matchlight Fingerprint
            Monitoring.
        records: :class:`~matchlight.record.RecordMethods` object with
            access to record methods for Matchlight Fingerprint
            Monitoring.
        feeds: :class:`~matchlight.feed.FeedMethods` object with access
            to all Matchlight DataFeed methods.
        search: :class:`~matchlight.search.SearchMethods` object with
            access to Matchlight Retrospective Search.

    """

    def __init__(self, access_key=None, secret_key=None, **kwargs):
        """Matchlight API connection object.

        This class is a wrapper over the Matchlight API feature set
        including functionality for search, feeds, projects, and
        records.

        Attributes:
            access_key (str, optional): The user's Matchlight API
                access key. If not passed as an argument this value
                must be set using the ``MATCHLIGHT_ACCESS_KEY``
                environment variable.
            secret_key (str, optional): The user's Matchlight API
                access key. If not passed as an argument this value
                must be set using the ``MATCHLIGHT_SECRET_KEY``
                environment variable.

        """
        self.conn = Connection(
            access_key=access_key, secret_key=secret_key, **kwargs)
        self.alerts = AlertMethods(self.conn)
        self.projects = ProjectMethods(self.conn)
        self.feeds = FeedMethods(self.conn)
        self.records = RecordMethods(self.conn)
        self.search = SearchMethods(self.conn).search
        self.pii_search = SearchMethods(self.conn).pii_search



'''
This module contains example functions demonstrating differeny ways of
filtering and pulling down Matchlight alerts data.
'''

import records
from matchlight import Matchlight

from local_alert_utilities import (
    get_last_saved_alert,
    get_number_of_local_alerts,
    save_new_alert,
    setup_example_database,
)


def get_and_store_alerts():
    # Create a database connection and set up a local alert database.
    # db = records.Database('sqlite:///matchlight-alerts-example.db')
    db = records.Database('sqlite:///:memory:')
    setup_example_database(db)
    print(
        f'Number of alerts in database is {get_number_of_local_alerts(db)}'
    )

    # Create a Matchliht API connection
    # Tip: You can generate your Matchlight API keys through the
    # Matchlight web interface
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html?highlight=keys#authentication
    ml = Matchlight(
        access_key='your-matchlight-api-access-key',
        secret_key='your-matchlight-api-secret-key'
    )

    # Grab a Matchlight project with multiple alerts.
    project = ml.projects.get('your-project-upload-token')

    # Uncomment to reset all you Matchlight data every time.
    # set_all_alerts_as_unseen(ml, project)

    last_alert = get_last_saved_alert(db)
    print(f'Latest alert is number {last_alert.max_alert_number}')

    keep_going = True
    while keep_going:
        alert_query = ml.alerts.filter(project=project, limit=1, seen=False)
        if len(alert_query) == 0:
            print('No unseen alerts left')
            break
        else:
            new_alert = alert_query[0]
        print(f'New Alert found with Alert Number {new_alert.number}')
        if new_alert.number > last_alert.max_alert_number:
            save_new_alert(db, new_alert)
            ml.alerts.edit(new_alert.id, seen=True)
            print(f'Alert Number {new_alert.number} set to "seen"')
        else:
            keep_going = False
            print('No new alerts remain')
    print(
        f'Number of alerts in databasse is {get_number_of_local_alerts(db)}'
    )


def set_all_alerts_as_unseen(ml, project):
    '''
    This is utility function that resets the last 50 alerts in a
    project to unseen. This had no application in production but is a
    development utility to reset test data.
    '''
    print('Resetting Alerts')
    last_fifty_alerts = ml.alerts.filter(project=project, limit=50)
    print(f'Found {len(last_fifty_alerts)} alerts')
    for alert in last_fifty_alerts:
        print(f'Setting alert number {alert.number} as seen.')
        ml.alerts.edit(alert.id, seen=False)


if __name__ == '__main__':
    get_and_store_alerts()



"""Manages a Matchlight API connection."""
from __future__ import absolute_import

import os

import requests
import requests.adapters
import requests.exceptions
import requests.packages.urllib3 as requests_urllib3

import matchlight.error


__all__ = (
    'Connection',
    'MATCHLIGHT_API_URL_V2',
)


MATCHLIGHT_API_URL_V2 = 'https://api.matchlig.ht/api/v2'


class Connection(object):
    """Matchlight API connection object."""

    def __init__(self, access_key=None, secret_key=None, https_proxy=None,
                 insecure=False, endpoint=None, search_endpoint=None):
        """Initializes a new API connection.

        Args:
            access_key (str, optitonal): The user's Matchlight Public
                API access key. If not passed as an argument this value
                must be set using the ``MATCHLIGHT_ACCESS_KEY``
                environment variable.
            secret_key (str, optional): The user's Matchlight Public
                API access key. If not passed as an argument this value
                must be set using the ``MATCHLIGHT_SECRET_KEY``
                environment variable.
            https_proxy (str): A string defining the HTTPS proxy to
                use. Defaults to None.
            insecure (bool, optional): Whether or not to verify
                certificates for the HTTPS proxy. Defaults to ``False``
                (certificates will be verified).
            endpoint (str, optional): Base URL for requests. Defaults
                to ``'https://api.matchlig.ht/api/v2'``.
            search_endpoint (str, optional): Base URL for all search
                API requests.

        """
        if access_key is None:
            access_key = os.environ.get('MATCHLIGHT_ACCESS_KEY', None)
        if secret_key is None:
            secret_key = os.environ.get('MATCHLIGHT_SECRET_KEY', None)
        if access_key is None or secret_key is None:
            raise matchlight.error.SDKError(
                'The APIConnection object requires your Matchlight '
                'API access_key and secret_key either be passed as input '
                'parameters or set in the MATCHLIGHT_ACCESS_KEY and '
                'MATCHLIGHT_SECRET_KEY environment variables.')
        if endpoint is None:
            endpoint = MATCHLIGHT_API_URL_V2
        if search_endpoint is None:
            search_endpoint = MATCHLIGHT_API_URL_V2

        self.access_key = access_key
        self.secret_key = secret_key
        self.proxy = {'https': https_proxy}
        self.insecure = insecure
        self.endpoint = endpoint
        self.search_endpoint = search_endpoint
        self.session = requests.Session()
        self.session.mount(
            self.endpoint,
            requests.adapters.HTTPAdapter(
                max_retries=requests_urllib3.util.Retry(
                    total=5, status_forcelist=[500, 502, 503, 504])),
        )

    def request(self, path, data=None, endpoint=None, **kwargs):
        """Send an HTTP request to the Matchlight API.

        Args:
            path (str): The path of request URL without the URL base.
                e.g. ``/search``.
            data (dict or list): Serializable data for ``POST``
                requests. Defaults to ``None``.
            endpoint (str, optional): option to pass a different endpoint for
                each request. Defaults to Connection().endpoint

        Returns:
            A :class:`requests.models.Response` object.

        Raises:
            ConnectionError: Raised when there is an ``RetryError`` or
                ``ConnectionError`` from ``requests``.
            APIError: Raised when API does not return a 200, including
                error and custom message, if available.

        """
        # Allows SDK to use different endponts for search
        if endpoint is None:
            endpoint = self.endpoint
        url = ''.join([endpoint, path])

        method = 'GET' if data is None else 'POST'
        if 'timeout' not in kwargs:
            # API timeout is 90s
            kwargs['timeout'] = 91.0

        response = self._request(
            method,
            url,
            data=data,
            headers={
                'Content-Type': 'application/json',
                'Matchlight-Request-Source': 'python-sdk',
                'Matchlight-Request-Source-Version': matchlight.__version__,
            },
            auth=(self.access_key, self.secret_key),
            proxies=self.proxy,
            verify=not self.insecure,
            **kwargs)
        return response

    def _request(self, method, url, data=None, **kwargs):
        try:
            response = self.session.request(method, url, data=data, **kwargs)
            if response.status_code == 200:
                return response
            else:
                try:
                    data = response.json()
                except ValueError:
                    data = None
                raise matchlight.error.APIError(
                    response.status_code, response.reason, data)
        except requests.exceptions.RetryError:
            raise matchlight.error.ConnectionError(
                'Matchlight API request failed with too many retries')
        except requests.exceptions.ConnectionError:
            raise matchlight.error.ConnectionError(
                'Matchlight API request failed with connection error')

    def __repr__(self):  # pragma: no cover
        return self.access_key



"""An interface for downloading and filtering Matchlight feeds."""
from __future__ import absolute_import, print_function

import datetime
import io
import json
import time

import six

import matchlight.error
import matchlight.utils

if not six.PY3:
    from backports import csv
else:
    import csv


class Feed(object):
    """Represents a Matchlight Data Feed.

    Examples:
            >>> ml = matchlight.Matchlight()
            >>> feed = ml.feeds.filter()[0]
            >>> feed
            <Feed(name="CompanyEmailAddress", recent_alerts=2)>
            >>> feed.details
            {'description': None, 'name': u'CompanyEmailAddress',
            'recent_alerts_count': 2,
            'start_timestamp': '2016-06-03T00:00:00',
            'stop_timestamp': None}

    Attributes:
        description (:obj:`str`): Description of the feed.
        name (:obj:`str`): Name of the feed.
        recent_alerts_count (int): Number of recent alerts.
        start_timestamp (:class:`datetime.datetime`): Start time of the
            feed.
        stop_timestamp (:class:`datetime.datetime`): Stop time of the
            feed.

    """

    def __init__(self, name, description, recent_alerts_count,
                 start_timestamp, stop_timestamp=None):
        """Initializes a new Matchlight feed.

        Args:
            description (:obj:`str`): Description of the feed.
            name (:obj:`str`): Name of the feed.
            recent_alerts_count (int): Number of recent alerts.
            start_timestamp (:obj:`str`): ISO 8601 formatted timestamp
                of when feed collection began.
            stop_timestamp (:obj:`str`, optional): ISO 8601 formatted
                timestamp of when feed collection ended or will end. If
                not provided, the feed is assumed to never expired.

        """
        self.name = name
        self.description = description
        self.recent_alerts_count = recent_alerts_count
        self.start_timestamp = start_timestamp
        self.stop_timestamp = stop_timestamp

    @property
    def details(self):
        """:obj:`dict`: Returns the feed details as a mapping."""
        return {
            'name': self.name,
            'description': self.description,
            'recent_alerts_count': self.recent_alerts_count,
            'start_timestamp': self.start_timestamp,
            'stop_timestamp': self.stop_timestamp,
        }

    @property
    def start(self):
        """:class:`datetime.datetime`: When feed data collection began."""
        return datetime.datetime.fromtimestamp(self.start_timestamp)

    @property
    def end(self):  # noqa: D205,D400
        """:obj:`NoneType` or :class:`datetime.datetime`: If the feed
        has a ``stop_timestamp``, returns a datetime object. Otherwise,
        returns :obj:`NoneType`.

        """
        if self.stop_timestamp:
            return datetime.datetime.fromtimestamp(self.stop_timestamp)

    def __repr__(self):  # pragma: no cover
        return '<Feed(name="{name}", recent_alerts={alerts})>'.format(
            alerts=self.recent_alerts_count, name=self.name)


class FeedMethods(object):
    """Provides methods for interfacing with the feeds API."""

    def __init__(self, ml_connection):  # noqa: D205,D400
        """Initializes a feed interface with the given Matchlight
        connection.

        Args:
            ml_connection (:class:`~.Connection`): A Matchlight
                connection instance.

        """
        self.conn = ml_connection

    def all(self):
        """Returns a list of feeds associated with a Matchlight account.

        Returns:
            :obj:`list` of :class:`matchlight.Feed`: A list of feeds
                associated with an account.

        """
        r = self.conn.request('/feeds')
        return [Feed(**feed) for feed in r.json()['feeds']]

    def counts(self, feed, start_date, end_date):
        """Daily counts for a feed for a given date range.

        Args:
            feed (:class:`~.Feed`): A feed instance or feed name.
            start_date (:class:`datetime.datetime`): Start of date range.
            end_date (:class:`datetime.datetime`): End of date range.

        Returns:
            :obj:`dict`: Mapping of dates (``YYYY-MM-DD``) to alert counts.

        """
        if isinstance(feed, six.string_types):
            feed_name = feed
        else:
            feed_name = feed.name

        data = {
            'start_date': int(matchlight.utils.datetime_to_unix(start_date)),
            'end_date': int(matchlight.utils.datetime_to_unix(end_date)),
        }
        response = self.conn.request(
            '/feeds/{feed_name}'.format(feed_name=feed_name),
            data=json.dumps(data))
        return self._format_count(response.json())

    def download(self, feed, start_date, end_date, save_path=None):
        """Downloads feed data for the given date range.

        Args:
            feed (:class:`~.Feed`): A feed instance or feed name.
            start_date (:class:`datetime.datetime`): Start of date range.
            end_date (:class:`datetime.datetime`): End of date range.
            save_path (:obj:`str`): Path to output file.

        Returns:
            :obj:`list` of :obj:`dict`: All feed hits for the given range.

        """
        if isinstance(feed, six.string_types):
            feed_name = feed
        else:
            feed_name = feed.name

        data = {
            'start_date': int(matchlight.utils.datetime_to_unix(start_date)),
            'end_date': int(matchlight.utils.datetime_to_unix(end_date)),
        }
        response = self.conn.request(
            '/feed/{feed_name}/prepare'.format(feed_name=feed_name),
            data=json.dumps(data))

        if response.status_code != 200:
            raise matchlight.error.SDKError(
                'Feed failed to be generated. Please try again later.')

        data = {'feed_response_id': response.json().get('feed_response_id')}

        status = 'pending'
        while status == 'pending':
            response = self.conn.request(
                '/feed/{feed_name}/link'.format(feed_name=feed_name),
                data=json.dumps(data))
            status = response.json().get('status', None)
            time.sleep(1)
            # TODO: backoff and timeout

        if status == 'failed':
            raise matchlight.error.SDKError(
                'Feed failed to be generated. Please try again later.')
        elif status == 'ready':
            content = self.conn._request('GET', response.json().get('url'))
        else:
            raise matchlight.error.SDKError('An unknown error occurred.')

        if save_path:
            with io.open(save_path, 'wb') as f:
                f.write(content.content)
        else:
            unicode_feed = content.content.decode('utf-8-sig')
            return [
                self._format_feed(row)
                for row in csv.DictReader(unicode_feed.split('\n'))
            ]

    def _format_count(self, counts):
        return {
            datetime.datetime.fromtimestamp(int(k)).strftime('%Y-%m-%d'): v
            for k, v in counts.items()
        }

    def _format_feed(self, feed_row):
        feed_row['ts'] = matchlight.utils.terbium_timestamp_to_datetime(
            feed_row['ts'])
        return feed_row

    def __iter__(self):
        return (item for item in self.all())



"""Test configuration for the Matchlight SDK."""
import time
import uuid

import pytest

import matchlight


@pytest.fixture
def access_key():
    """Provides a fake access key in the form of a UUID4."""
    return uuid.uuid4().hex


@pytest.fixture
def secret_key():
    """Provides a fake secret key in the form of a UUID4."""
    return uuid.uuid4().hex


@pytest.fixture
def connection(access_key, secret_key):
    """A connection object initialized with a fake access and secret key."""
    return matchlight.Matchlight(access_key=access_key, secret_key=secret_key)


@pytest.fixture(scope='function')
def id():
    """Provides a fake id in the form of a UUID4."""
    return str(uuid.uuid4())


@pytest.fixture(scope='function')
def number_of_records():
    """A record count fixture."""
    return 10


@pytest.fixture(scope='function')
def number_of_unseen_alerts():
    """An unseen alerts count fixture."""
    return 10


@pytest.fixture(scope='function')
def document(request):
    """A document mapping fixture."""
    return {
        'id': uuid.uuid4().hex,
        'name': 'Document record',
        'description': '',
        'ctime': time.time(),
        'mtime': time.time(),
        'metadata': {},
    }


@pytest.fixture(scope='function')
def document_record(document):
    """A document record fixture."""
    return matchlight.Record(**document)


@pytest.fixture(scope='function')
def project_name():
    """A project name fixture."""
    return 'Test Project'


@pytest.fixture(scope='function', params=[
    'bulk_pii',
    'document',
    'pii',
    'source_code',
])
def project_type(request):
    """A parametrized project type fixture."""
    return request.param


@pytest.fixture(scope='function')
def upload_token():
    """An upload token (dash-separated UUID4) fixture."""
    return str(uuid.uuid4())


@pytest.fixture(scope='function')
def project_payload(project_name, project_type, upload_token,
                    number_of_records, number_of_unseen_alerts):
    """A project payload artifact, parametrized by project type."""
    return {
        'project_name': project_name,
        'project_type': project_type,
        'project_upload_token': upload_token,
        'last_date_modified': time.time(),
        'number_of_records': number_of_records,
        'number_of_unseen_alerts': number_of_unseen_alerts,
    }


@pytest.fixture
def project(project_payload):
    """A project instance fixture, parametrized by project type."""
    return matchlight.Project.from_mapping(project_payload)


@pytest.fixture(scope='function')
def alert_payload(id, upload_token):
    """An alert payload artifact."""
    return {
        'id': id,
        'alert_number': 10,
        'type': 'pii',
        'url': 'https://terbiumlabs.com/matchlight.html',
        'url_metadata': {
            'description': 'Matchlight provides intelligence on your most imp',
            'tor_only': 'false'
        },
        'ctime': time.time(),
        'mtime': time.time(),
        'seen': 'true',
        'archived': 'true',
        'upload_token': upload_token,
        'details': {
            'pii': {
                'fields': ['phone']
            }
        },
        'asset_name': 'Example Record',
        'project_name': 'Sample Project',
    }


@pytest.fixture(scope='function')
def alert_details_pii_payload():
    """An alert details payload artifact for a pii alert."""
    return {
        'details': {
            'pii': [
                {
                    'email': 'o****@gmail.com',
                    'first': 'a****',
                    'last': 'b****',
                    'record_id': 'd3c59d38c4054f62876a2a7a3dca41ca'
                }
            ]
        },
        'notes': '',
        'type': 'pii'
    }


@pytest.fixture
def alert(alert_payload):
    """An alert instance fixture."""
    return matchlight.Alert.from_mapping(alert_payload)


@pytest.fixture(scope='function')
def pii_search_email_only_results():
    """PII search results for only the email field."""
    return [
        {
            'fields': ['email'],
            'ts': '2018-07-25T20:00:44',
            'source': 'Exactis Breach June 2018'
        },
        {
            'fields': ['email'],
            'ts': '2017-01-25T02:35:04',
            'source': 'https://pastebin.com/raw.php?i=1DgbtSZc'
        },
        {
            'fields': ['email'],
            'ts': 1556221970,
            'source': 'Zoosk Breach Nov 2016'
        },
        {
            'fields': ['email'],
            'ts': '1558333205',
            'source': 'https://www.reddit.com/r/AskReddit/comments/3oqj4a'
        },
    ]


@pytest.fixture(scope='function')
def search_email_only_results():
    """PII search results for only the email field."""
    return [
        {
            'cwid': 'fff33cbe7ed54f5ebfccd09c3d24999c',
            'score': 800,
            'ts': 1453298293,
            'urls': [
                [
                    1453298293,
                    (
                        'http://blockchainbdgpzk.onion/tx/4f4097992b89156'
                        '690817556fc3f540535bdfadde06661c9cae21d500943f970'
                    )
                ]
            ]
        },
        {
            'cwid': 'ffedf0a2b775cfd50383603ff827d702',
            'score': 400,
            'ts': 1438870917,
            'urls': [
                [
                    1438870917,
                    (
                        'http://nqigfqrxnkwcqmiq.onion/'
                        'wiki/index.php#Whistleblowing'
                    )
                ]
            ]
        },
    ]



"""Exceptions raised by the Matchlight SDK."""
__all__ = (
    'APIError',
    'ConnectionError',
    'InvalidCredentialsError',
    'SDKError',
)


class SDKError(Exception):
    """Errors originating from within the Matchlight SDK."""


class APIError(SDKError):
    """Errors passed through from the Matchlight API."""


class ConnectionError(SDKError):
    """Custom error class for API request failures."""


class InvalidCredentialsError(ConnectionError):
    """Exception thrown when 401 Forbidden occurs."""



"""Unit tests the utils module of Matchlight SDK."""
import pytest

import matchlight


@pytest.mark.parametrize('test_input,expected', [
    (None, '*****'),
    ('', '*****'),
    ('a', 'a****'),
    ('foobar', 'f****'),
    ('terbiumlabs', 't****'),
])
def test_blind_name(test_input, expected):
    """Verify test cases for blind_name."""
    assert matchlight.utils.blind_name(test_input) == expected


@pytest.mark.parametrize('test_input,expected', [
    (None, '*****'),
    ('', '*****'),
    ('a@terbiumlabs.com', 'a****@terbiumlabs.com'),
    ('ab@terbiumlabs.com', 'a****@terbiumlabs.com'),
    ('abc@terbiumlabs.com', 'a****@terbiumlabs.com'),
    ('abcd@terbiumlabs.com', 'ab****@terbiumlabs.com'),
    ('abcde@terbiumlabs.com', 'ab****@terbiumlabs.com'),
    ('abcdef@terbiumlabs.com', 'abc****@terbiumlabs.com'),
    ('abcdefg@terbiumlabs.com', 'abc****@terbiumlabs.com'),
    ('abcdefgh@terbiumlabs.com', 'abc****@terbiumlabs.com'),
])
def test_blind_email(test_input, expected):
    """Verify test cases for blind_email."""
    assert matchlight.utils.blind_email(test_input) == expected



"""An interface for creating and retrieving alerts in Matchlight."""
from __future__ import absolute_import

import calendar
import datetime
import json

import matchlight.error


__all__ = (
    'Alert',
    'AlertMethods',
)


class Alert(object):
    """Represents an alert.

    Attributes:
        id (:obj:`str`): A 128-bit UUID.
        number (:obj:`int`): The account specific alert number.
        type (:obj:`str`): The type of the associated Record.
        url (:obj:`str`): The url where the match was found.
        url_metadata (:obj:`dict`): additional information about the url.
        ctime (:obj:`int`): A Unix timestamp of the alert creation
            timestamp.
        mtime (:obj:`int`): A Unix timestamp of the alert last
            modification date timestamp.
        seen (:obj:`bool`): User specific flag.
        archived (:obj:`bool`): User specific flag.
        upload_token (:obj:`str`): The upload_token of the associated Project.
        details (:obj:`dict`): Additional information about the Alert.
        project_name (:obj:`str`): The name of the associated Project
        record_name (:obj:`str`): The name of the associated Record.

    """

    def __init__(self, id, number, type, url, url_metadata, ctime, mtime, seen,
                 archived, upload_token, details, project_name, record_name):
        """Initializes a new alert.

        Args:
            id (:obj:`str`): A 128-bit UUID.
            number (:obj:`int`): The account specific alert number.
            type (:obj:`str`): The type of the associated Record.
            url (:obj:`str`): The url where the match was found.
            ctime (:obj:`int`): A Unix timestamp of the
                alert creation timestamp.
            mtime (:obj:`int`): A Unix timestamp of the
                alert last modification date timestamp.
            seen (:obj:`bool`): User specific flag.
            archived (:obj:`bool`): User specific flag.
            upload_token (:obj:`str`): The upload_token of the associated
                record.
            details (:obj:`dict`): details about the Alert.
            project_name (:obj:`str`): The name of the associated Project
            record_name (:obj:`str`): The name of the associated Record.

        """
        self.id = id
        self.number = number
        self.type = type
        self.url = url
        self.url_metadata = url_metadata
        self.ctime = ctime
        self.mtime = mtime
        self.seen = seen
        self.archived = archived
        self.upload_token = upload_token
        self.details = details
        self.project_name = project_name
        self.record_name = record_name

    @classmethod
    def from_mapping(cls, mapping):
        """Creates a new alert instance from the given mapping."""
        return cls(
            id=mapping['id'],
            number=mapping['alert_number'],
            type=mapping['type'],
            url=mapping['url'],
            url_metadata=mapping['url_metadata'],
            ctime=mapping['ctime'],
            mtime=mapping['mtime'],
            seen=True if mapping['seen'] == 'true' else False,
            archived=True if mapping['archived'] == 'true' else False,
            upload_token=mapping['upload_token'],
            details=mapping['details'],
            project_name=mapping['project_name'],
            record_name=mapping['asset_name']
        )

    @property
    def last_modified(self):
        """:class:`datetime.datetime`: The last modified timestamp."""
        if self.mtime is None:
            return None
        return datetime.datetime.fromtimestamp(self.mtime)

    @property
    def date(self):
        """:class:`datetime.datetime`: The date created timestamp."""
        if self.ctime is None:
            return None
        return datetime.datetime.fromtimestamp(self.ctime)

    @property
    def score(self):
        """:obj:`int`: Represents how much of the record appeared on the page.

        Scores range from 1 to 800, with 800 representing that the entire
        record was found on the page. PII records will always have a score
        of 800.
        """
        if self.type == 'pii':
            return 800
        if self.type == 'document':
            return self.details['document'].get('score', None)
        if self.type == 'source_code':
            return self.details['source_code'].get('score', None)

    @property
    def fields(self):
        """:obj:`list`: PII records will match on one or more 'fields'."""
        if self.type == 'pii':
            return self.details['pii'].get('fields', [])

        return None

    def __repr__(self):  # pragma: no cover
        return '<Alert(number="{}", id="{}")>'.format(
            self.number,
            self.id
        )


class AlertMethods(object):
    """Provides methods for interfacing with the alerts API."""

    def __init__(self, ml_connection):  # noqa: D205,D400
        """Initializes an alerts interface with the given Matchlight
        connection.

        Args:
            ml_connection (:class:`~.Connection`): A Matchlight
                connection instance.

        """
        self.conn = ml_connection

    def filter(self, limit, seen=None, archived=None, project=None,
               record=None, last_modified=None, last_alert=None):
        """Returns a list of alerts.

        Providing a **limit** keyword argument will limit the number of alerts
        returned. The request may time out if this is set too high, a limit of
        50 is recomended to avoid timeouts.

        Providing an optional **seen** keyword argument will only
        return alerts that match that property

        Providing an optional **archived** keyword argument will only
        return alerts that match that property

        Providing an optional **project** keyword argument will only
        return alerts that are associated with a specific project.

        Providing an optional **record** keyword argument will only
        return alerts that are associated with a specific record.

        Providing an optional **last_modified** keyword argument will only
        return alerts with a last_modifed less than the argument.

        Providing an optional **last_alert** keyword argument will only return
        results after this Alert

        Examples:
            Request all unseen alerts::

                >>> ml.alerts.filter(seen=False, limit=50)
                [<Alert(number="1024",
                id="625a732ad0f247beab18595z951c2088a3")>,
                Alert(number="1025",
                id="f9427dd5a24d4a98b2069004g04c2977")]

            Request all alerts for a project::

                >>> my_project
                <Project(name="Super Secret Algorithm", type="source_code")>
                >>> ml.alerts.filter(project=my_project, limit=50)
                [<Alert(number="1024",
                id="625a732ad0f247beab18595z951c2088a3")>,
                Alert(number="1025",
                id="f9427dd5a24d4a98b2069004g04c2977")]

            Request sets of alerts using pagination::

                >>> ml.alerts.filter(limit=50)
                [<Alert(number="1027",
                id="625a732ad247beab18595z951c2088a3")>,
                Alert(number="1026",
                id="f9427dd5a24d4a98b2069004g04c2977")...
                >>> ml.alerts.filter(limit=50, last_alert=50)
                [<Alert(number="977",
                id="59d5a791g8d4436aaffe64e4b15474a5")>,
                Alert(number="976",
                id="6b1001aaec5a48f19d17171169eebb56")...

        Args:
            limit (:obj:`int`):
                Don't return more than this number of alerts.
            seen (:obj:`bool`, optional):
            archived (:obj:`bool`, optional):
            project (:class:`~.Project`, optional): a project object.
                Defaults to all projects if not specified.
            record (:class:`~.Record`, optional): a record object.
                Defaults to all projects if not specified.
            last_modified (:obj:`datetime`, optional):
            last_alert (:obj:`int`):
                Only return Alerts after this one.

        Returns:
            :obj:`list` of :class:`~.Alert`: List of alerts that
                are associated with a project.

        """
        if seen is not None:
            seen_int = 1 if seen is True else 0
        else:
            seen_int = None

        if archived is not None:
            archived_int = 1 if archived is True else 0
        else:
            archived_int = None

        if project is not None:
            upload_token = project.upload_token
        else:
            upload_token = None

        if record is not None:
            record_id = record.id
        else:
            record_id = None

        if last_modified is not None:
            mtime = calendar.timegm(last_modified.timetuple())
        else:
            mtime = None

        response = self.conn.request(
            '/alerts',
            params={
                'limit': limit,
                'seen': seen_int,
                'archived': archived_int,
                'upload_token_filter': upload_token,
                'record_id_filter': record_id,
                'mtime': mtime,
                'last_alert': last_alert
            }
        )
        alerts = []
        for payload in response.json().get('alerts', []):
            alerts.append(Alert.from_mapping(payload))
        return alerts

    def edit(self, alert_id, seen=None, archived=None):
        """Edits an alert.

        Example:
            Archive an alert::

                >>> alert
                <Alert(number=1024,
                id="0760570a2c4a4ea68d526f58bab46cbd")>
                >>> ml.alerts.edit(alert, archived=True)
                {
                    'seen': True,
                    'archived': True
                }

        Arguments:
            alert (:obj:`str`): An alert id.
            seen (:obj:`bool`, optional):
            archived (:obj:`bool`, optional):

        Returns:
            :obj:`dict`: Updated alert metadata.

        """
        if isinstance(alert_id, Alert):
            alert_id = alert_id.id

        data = {}
        if seen is not None:
            data['seen'] = seen

        if archived is not None:
            data['archived'] = archived

        response = self.conn.request(
            '/alert/{}/edit'.format(alert_id),
            data=json.dumps(data)
        )
        response = response.json()
        return {
            'seen': response['seen'],
            'archived': response['archived']
        }

    def get_details(self, alert_id):
        """Returns details of an alert by the given alert ID.

        Args:
            alert_id (:obj:`str`): The alert identifier.

        Returns:
           :obj:`dict`: map of the alert details.

        """
        if isinstance(alert_id, Alert):
            alert_id = alert_id.id

        try:
            response = self.conn.request('/alert/{}/details'.format(alert_id))
            return response.json()
        except matchlight.error.APIError as err:
            if err.args[0] == 404:
                return
            else:
                raise



"""Unit tests the record methods of the Matchlight SDK."""
from __future__ import unicode_literals

import io
import json
import random
import string
import time
import uuid

import pytest
import responses

import matchlight


DOCUMENT_RECORD_PATH = 'tests/fixtures/UTF-8-test.txt'
PII_RECORDS_RAW_PATH = 'tests/fixtures/pii_records_raw.json'


# fake = faker.Factory.create()


# See <https://github.com/pytest-dev/pytest-factoryboy/issues/29>
# class PiiRecord(object):
#
#     def __init__(self, email, first_name, middle_name, last_name,
#                  ssn, address, city, state, zipcode, phone):
#         self.email = email
#         self.first_name = first_name
#         self.middle_name = middle_name
#         self.last_name = last_name
#         self.ssn = ssn
#         self.address = address
#         self.city = city
#         self.state = state
#         self.zipcode = zipcode
#         self.phone = phone
#
#
# @pytest_factoryboy.register
# class PiiRecordFactory(factory.Factory):
#
#     class Meta:
#         model = PiiRecord
#
#     email = factory.LazyAttribute(lambda _: fake.email())
#     first_name = factory.LazyAttribute(lambda _: fake.first_name())
#     middle_name = factory.LazyAttribute(lambda _: fake.first_name())
#     last_name = factory.LazyAttribute(lambda _: fake.last_name())
#     ssn = factory.LazyAttribute(lambda _: fake.ssn())
#     address = factory.LazyAttribute(lambda _: fake.address())
#     city = factory.LazyAttribute(lambda _: fake.city())
#     state = factory.LazyAttribute(lambda _: fake.state())
#     zipcode = factory.LazyAttribute(lambda _: fake.zipcode())
#     phone = factory.LazyAttribute(lambda _: fake.phone_number())


@pytest.fixture(scope='function', params=[PII_RECORDS_RAW_PATH])
def pii_records_raw(request):
    """Sample PII records."""
    with io.open(request.param) as fp:
        return json.loads(fp.read())


def test_record_details(document):
    """Verifies that record details are returned correctly."""
    record = matchlight.Record(**document)
    assert record.details == document


def test_record_user_provided_id(document):
    """Verifies record user provided identifiers."""
    document['metadata']['user_record_id'] = 42
    record = matchlight.Record(**document)
    assert record.user_provided_id == document['metadata']['user_record_id']


def test_record_add_pii_missing_first_name_error(connection, project):
    """Test that add_pii throws an error for missing first_name argument."""
    missing_first_name_record_data = {
        'email': 'heather03@hotmail.com',
        'middle_name': '',
        'last_name': 'whatever',
        'ssn': '000-00-0000',
        'address': '25068 Treena Cliff',
        'city': 'Almediashire',
        'state': 'Georgia',
        'zipcode': 82360,
        'phone': '04962816540'
    }

    with pytest.raises(matchlight.error.SDKError) as err_info:
        connection.records.add_pii(
            project=project,
            description='',
            **missing_first_name_record_data
        )

    assert str(err_info.value) == (
        'Fingerprinter Failed: the first_name argument is required along with '
        'the last_name argument.'
    )


def test_record_add_pii_missing_last_name_error(connection, project):
    """Test that add_pii throws an error for missing last_name argument."""
    missing_last_name_record_data = {
        'email': 'heather03@hotmail.com',
        'first_name': 'Elby',
        'middle_name': '',
        'ssn': '000-00-0000',
        'address': '25068 Treena Cliff',
        'city': 'Almediashire',
        'state': 'Georgia',
        'zipcode': 82360,
        'phone': '04962816540'
    }

    with pytest.raises(matchlight.error.SDKError) as err_info:
        connection.records.add_pii(
            project=project,
            description='',
            **missing_last_name_record_data
        )

    assert str(err_info.value) == (
        'Fingerprinter Failed: the last_name argument is required along with '
        'the first_name argument.'
    )


@responses.activate
@pytest.mark.parametrize('min_score', [
    None,
    800,
])
def test_record_add_document_prototype(
        min_score, connection, project, document):
    """Verifies adding document records to a project."""
    with io.open(DOCUMENT_RECORD_PATH, 'rb') as f:
        content = f.readline()[0:840]

    responses.add(
        method=responses.POST,
        url='{}/records/upload/document/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        json={
            'id': uuid.uuid4().hex,
            'name': 'name',
            'description': '',
            'ctime': time.time(),
            'mtime': time.time(),
            'metadata': '{}',
        },
        status=200
    )

    record = connection.records.add_document(
        project=project,
        name=document['name'],
        description=document['description'],
        content=content,
        user_record_id='12345',
        min_score=min_score)
    assert len(responses.calls) == 1

    record = connection.records.add_document(
        project=project,
        name=document['name'],
        description=document['description'],
        content=content,
        user_record_id=12345,
        min_score=min_score,
        offline=True)
    assert isinstance(record, dict)
    assert len(responses.calls) == 1

    # Test that a nice error is returned for too long input
    with pytest.raises(matchlight.error.SDKError) as err_info:
        connection.records.add_document(
            project=project,
            name=document['name'],
            description=document['description'],
            content=''.join(random.choice(
                string.ascii_uppercase + string.digits
            ) for _ in range(860)),
        )

    assert str(err_info.value) == (
        'Fingerprinter Failed: the maximum length of a Document record'
        ' is 840 characters.'
    )


@responses.activate
@pytest.mark.parametrize('min_score', [
    None,
    800,
])
def test_record_add_document(min_score, connection, project, document):
    """Verifies adding document records to a project."""
    with io.open(DOCUMENT_RECORD_PATH, 'rb') as f:
        content = f.readline()[0:840]

    responses.add(
        method=responses.POST,
        url='{}/records/upload/document/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        json={
            'id': uuid.uuid4().hex,
            'name': 'name',
            'description': '',
            'ctime': time.time(),
            'mtime': time.time(),
            'metadata': '{}',
        },
        status=200
    )

    record = connection.records.add_document(
        project=project,
        name=document['name'],
        description=document['description'],
        content=content,
        user_record_id='12345',
        min_score=min_score)
    assert len(responses.calls) == 1

    record = connection.records.add_document(
        project=project,
        name=document['name'],
        description=document['description'],
        content=content,
        user_record_id=12345,
        min_score=min_score,
        offline=True)
    assert isinstance(record, dict)
    assert len(responses.calls) == 1


@responses.activate
def test_record_add_pii(connection, project, pii_records_raw):
    """Verifies adding PII records to a project."""
    record_data = [
        {
            'id': uuid.uuid4().hex,
            'name': matchlight.utils.blind_email(record['email']),
            'description': '',
            'ctime': time.time(),
            'mtime': time.time(),
        }
        for record in pii_records_raw
    ]

    for payload in record_data:
        responses.add(
            method=responses.POST,
            url='{}/records/upload/pii/{}'.format(
                matchlight.MATCHLIGHT_API_URL_V2,
                project.upload_token
            ),
            json={
                'id': payload['id'],
                'name': payload['name'],
                'description': payload['description'],
                'ctime': payload['ctime'],
                'mtime': payload['mtime'],
                'metadata': '{}',
            },
            status=200
        )

    for counter, pii_record in enumerate(pii_records_raw):
        record = connection.records.add_pii(
            project=project,
            description='',
            **pii_record)
        assert record.id == record_data[counter]['id']
        assert len(responses.calls) == counter + 1

    for counter, pii_record in enumerate(pii_records_raw):
        record = connection.records.add_pii(
            project=project,
            description='',
            offline=True,
            **pii_record)
        assert isinstance(record, dict)
        assert len(responses.calls) == 6



'''
This module contains an example script showing how to upload PII
data using the Matchlight SDK.
'''

import csv
import uuid

from matchlight import Matchlight


def get_pii_data():
    '''Get the plain text PII data to fingerprint.

    This is just a placeholder for any function that loads plaintext
    data, in practice this could pull data in from anywhere: a
    database, a CSV, another API, etc.
    '''
    return [
        {
            'email': 'richardwoodward@gardner.biz',
            'first_name': 'Preston',
            'last_name': 'Wiggins',
            'address': '83037 Amy Lodge',
            'city': 'Alexanderstad',
            'state': 'New Mexico',
            'zipcode': '81075',
            'phone': '1035223952',
            'ssn': '129602039',
            'credit_card': '4916528483361601'
        },
        {
            'email': 'daniel58@gmail.com',
            'first_name': 'Edward',
            'last_name': 'Figueroa',
            'address': 'Unit 1041 Box 1550',
            'city': 'Brendamouth',
            'state': 'South Carolina',
            'zipcode': '07721',
            'phone': '(434)191-8466',
            'ssn': '096571603',
            'credit_card': '4532566428800162'
        },
        {
            'email': 'gregory39@schultz.com',
            'first_name': 'Hunter',
            'last_name': 'Hill',
            'address': 'PSC 5552, Box 2365',
            'city': 'Martinshire',
            'state': 'Mississippi,67262',
            'zipcode': '32334',
            'phone': '9338226',
            'ssn': '261500223',
            'credit_card': '5569181104928858'
        }
    ]


def save_formatted_data(formatted_data):
    with open('formatted-data.csv', 'w', newline='') as csvfile:
        writer = csv.DictWriter(
            csvfile, fieldnames=formatted_data[0].keys()
        )
        writer.writeheader()
        for item in formatted_data:
            writer.writerow(item)


def main():
    '''Fingerprint PII data and upload to Matchlight.'''
    plaintext_pii_data = get_pii_data()

    # Add a unique `user_record_id` to each record as an index to
    # connect plaintext data back to Matchlight fingerprints.
    for record in plaintext_pii_data:
        record.update({'user_record_id': uuid.uuid4()})

    # Tip: You should save this formatted plaintext output so you will
    # be able to connect Matchlight alerts back to source input data
    # using the `user_record_id` field.
    save_formatted_data(plaintext_pii_data)

    # Tip: You can generate your Matchlight API keys through the
    # Matchlight web interface
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html?highlight=keys#authentication
    ml = Matchlight(
        access_key='your-matchlight-api-access-key',
        secret_key='your-matchlight-api-secret-key'
    )
    # Tip: Make sure you have already created a project
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html#create-a-new-project
    project = ml.projects.get('your-project-upload-token')

    total_records = len(plaintext_pii_data)
    for counter, record in enumerate(plaintext_pii_data):
        print(
            f'Uploading {record["email"]} '
            f'({counter + 1} of {total_records})'
        )
        ml.records.add_pii(
            project=project,
            description='Created from SDK example script',
            email=record['email'],
            first_name=record['first_name'],
            middle_name=None,
            last_name=record['last_name'],
            ssn=record['ssn'],
            address=record['address'],
            city=record['city'],
            state=record['state'],
            zipcode=record['zipcode'],
            phone=record['phone'],
            credit_card=record['credit_card'],
            user_record_id=str(record['user_record_id'])
        )


if __name__ == '__main__':
    main()



from ctypes import cdll, c_char, c_char_p, c_int, create_string_buffer, Structure, POINTER, pointer
from os import path
import platform
import sys
import os
import json
import pkg_resources

class LibfpException(Exception):
    pass

def get_library_path(library_filename):
    return pkg_resources.resource_filename('pylibfp', path.join('lib', library_filename))

_ISPY2 = sys.version_info[0] == 2
_ISPY3 = sys.version_info[0] == 3

_ARCH, _PLAT = platform.architecture()
if _PLAT.upper().startswith("WINDOWS"):
    if _ARCH.startswith("32"):
        _libfp = cdll.LoadLibrary(get_library_path("libfpw32.dll"))
    else:
        _libfp = cdll.LoadLibrary(get_library_path("libfpw64.dll"))
elif platform.system().upper() == "DARWIN":
    _libfp = cdll.LoadLibrary(get_library_path("libfpd64.so"))
else:
    _libfp = cdll.LoadLibrary(get_library_path("libfpl64.so"))


class ResultStruct(Structure):
  _fields_ = [("err", c_int), ("data", c_char_p)]


_libfp.fingerprint.argtypes = [POINTER(c_char), c_int, c_int]
_libfp.fingerprint.restype = POINTER(ResultStruct)

_libfp.fingerprint_chunk.argtypes = [POINTER(c_char), c_int]
_libfp.fingerprint_chunk.restype = c_char_p

_libfp.fingerprint_email_terbiumtwox.argtypes = [POINTER(c_char)]
_libfp.fingerprint_email_terbiumtwox.restype = c_char_p

_libfp.fingerprint_email_fromsha1.argtypes = [POINTER(c_char)]
_libfp.fingerprint_email_fromsha1.restype = c_char_p

_libfp.fingerprint_email_sha1.argtypes = [POINTER(c_char)]
_libfp.fingerprint_email_sha1.restype = c_char_p

_libfp.clean_string.argtypes = [POINTER(c_char), c_int]
_libfp.clean_string.restype = c_char_p

_libfp.assets_from_name.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_name.restype = c_char_p

_libfp.assets_from_address.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char)]
_libfp.assets_from_address.restype = c_char_p

_libfp.assets_from_city_state_zip.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_city_state_zip.restype = c_char_p

_libfp.assets_from_email_address.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char)]
_libfp.assets_from_email_address.restype = c_char_p

_libfp.assets_from_ssn.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char)]
_libfp.assets_from_ssn.restype = c_char_p

_libfp.assets_from_phone_number.argtypes = [POINTER(c_char), POINTER(c_char),
    POINTER(c_char)]
_libfp.assets_from_phone_number.restype = c_char_p

_libfp.assets_from_credit_card.argtypes = [POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_credit_card.restype = c_char_p

_libfp.assets_from_medicare_id.argtypes = [POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_medicare_id.restype = c_char_p

_libfp.assets_from_passport.argtypes = [POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_passport.restype = c_char_p

_libfp.assets_from_iban.argtypes = [POINTER(c_char), POINTER(c_char), POINTER(c_char)]
_libfp.assets_from_iban.restype = c_char_p

MODE_TEXT = 0
MODE_CODE = 1
MODE_DIGITS = 2

ML_HASH_LENGTH = 32

OPTIONS_BOOLEAN = 1
OPTIONS_RAW = 2
OPTIONS_TILED = 4
OPTIONS_STORABLENGRAMS = 8
OPTIONS_PARSEONLY = 16

ERROR_PADDING = 1
ERROR_PADDING_MSG = "Unable to fingerprint chunk: {}"


def _ensure_bytes(v):
    if _ISPY3 and hasattr(v, "encode"):
        return v.encode("utf-8")
    if _ISPY2 and type(v) is not str:
        try:
            return v.encode("utf-8")
        except AttributeError:
            return v
    else:
        return v


def _ensure_unicode(v):
    if _ISPY3 and hasattr(v, "decode"):
        return v.decode("utf-8")
    if _ISPY2 and type(v) is str:
        return v.decode("utf-8")
    else:
        return v


def _decode_raw(s):
    c = ML_HASH_LENGTH * 2
    return [_ensure_unicode(s[i:i+c]) for i in range(0, len(s), c)]


def fingerprint(content, mode=MODE_TEXT, boolean=False, raw=False, flags=0):
    result = None
    opts = 0
    if boolean and raw:
        raise ValueError("boolean and raw cannot both be true")
    if boolean:
        opts = opts | OPTIONS_BOOLEAN
        opts = opts | OPTIONS_TILED
    if raw:
        opts = opts | OPTIONS_RAW
    else:
        opts = opts | OPTIONS_STORABLENGRAMS
    opts = opts | flags
    result = _libfp.fingerprint(create_string_buffer(_ensure_bytes(content)), mode, opts)
    if result.contents.err == ERROR_PADDING:
        raise LibfpException(ERROR_PADDING_MSG.format(result.contents.data))
    if raw:
        return _decode_raw(result.contents.data)
    return _ensure_unicode(result.contents.data)


def fingerprint_chunk(content, flags=0):
    result = _libfp.fingerprint_chunk(create_string_buffer(_ensure_bytes(content)), flags)
    return _ensure_unicode(result)


def fingerprint_email_terbiumtwox(content):
    result = _libfp.fingerprint_email_terbiumtwox(create_string_buffer(_ensure_bytes(content)))
    return _ensure_unicode(result)


def fingerprint_email_fromsha1(content):
    result = _libfp.fingerprint_email_fromsha1(create_string_buffer(_ensure_bytes(content)))
    return _ensure_unicode(result)


def fingerprint_email_sha1(content):
    result = _libfp.fingerprint_email_sha1(create_string_buffer(_ensure_bytes(content)))
    return _ensure_unicode(result)


def clean_string(content, mode=MODE_TEXT):
    v = _ensure_bytes(content)
    return _ensure_unicode(_libfp.clean_string(v, mode))

def fingerprints_pii_name_variants(first_name, middle_name, last_name):
    first_name = _ensure_bytes(first_name)
    if middle_name is not None:
        middle_name = _ensure_bytes(middle_name)
    last_name = _ensure_bytes(last_name)
    assets_json = _ensure_unicode(_libfp.assets_from_name(
            b"temporary", b"0", first_name, middle_name, last_name))
    assets = json.loads(assets_json)
    return [[_ensure_unicode(fp) for fp in asset["fingerprints"]]
            for asset in assets]

def fingerprints_pii_address_variants(street_address):
    street_address = _ensure_bytes(street_address)
    assets_json = _ensure_unicode(_libfp.assets_from_address(
            b"temporary", b"0", street_address))
    assets = json.loads(assets_json)
    return [[_ensure_unicode(fp) for fp in asset["fingerprints"]]
            for asset in assets]

def fingerprints_pii_city_state_zip_variants(city, state, zipcode):
    city = _ensure_bytes(city)
    state = _ensure_bytes(state)
    zipcode = _ensure_bytes(zipcode)

    # MdY: This is required to make state optional, not sure why.
    state = b'' if state is None else state

    assets_json = _ensure_unicode(_libfp.assets_from_city_state_zip(
            b"temporary", b"0", city, state, zipcode))
    assets = json.loads(assets_json)
    return [[_ensure_unicode(fp) for fp in asset["fingerprints"]]
            for asset in assets]

def fingerprints_pii_email_address(email):
    email = _ensure_bytes(email)
    assets_json = _ensure_unicode(_libfp.assets_from_email_address(
            b"temporary", b"0", email))
    assets = json.loads(assets_json)
    return [
        [_ensure_unicode(fp) for fp in asset["fingerprints"]] for asset in assets
    ]

def fingerprints_pii_ssn(ssn):
    ssn = _ensure_bytes(ssn)
    assets_json = _ensure_unicode(_libfp.assets_from_ssn(
            b"temporary", b"0", ssn))
    assets = json.loads(assets_json)
    assert(len(assets) == 1)
    return [_ensure_unicode(fp) for fp in assets[0]["fingerprints"]]

def fingerprints_pii_phone_number(phone_number):
    phone_number = _ensure_bytes(phone_number)
    assets_json = _ensure_unicode(_libfp.assets_from_phone_number(
            b"temporary", b"0", phone_number))
    assets = json.loads(assets_json)
    assert(len(assets) == 1)
    return [_ensure_unicode(fp) for fp in assets[0]["fingerprints"]]

def fingerprints_pii_credit_card(credit_card):
    credit_card = _ensure_bytes(credit_card)
    assets_json = _ensure_unicode(_libfp.assets_from_credit_card(b"temporary", b"0", credit_card))
    assets = json.loads(assets_json)
    assert(len(assets) == 1)
    return [_ensure_unicode(fp) for fp in assets[0]["fingerprints"]]

def fingerprints_pii_medicare_id(medicare_id):
    medicare_id = _ensure_bytes(medicare_id)
    assets_json = _ensure_unicode(_libfp.assets_from_medicare_id(b"temporary", b"0", medicare_id))
    assets = json.loads(assets_json)
    assert(len(assets) == 1)
    return [_ensure_unicode(fp) for fp in assets[0]["fingerprints"]]

def fingerprints_pii_passport(passport):
    passport = _ensure_bytes(passport)
    assets_json = _ensure_unicode(_libfp.assets_from_passport(b"temporary", b"0", passport))
    assets = json.loads(assets_json)
    assert(len(assets) == 1)
    return [_ensure_unicode(fp) for fp in assets[0]["fingerprints"]]

def fingerprints_pii_iban(iban):
    iban = _ensure_bytes(iban)
    assets_json = _ensure_unicode(_libfp.assets_from_iban(b"temporary", b"0", iban))
    assets = json.loads(assets_json)

    return [[_ensure_unicode(fp) for fp in asset["fingerprints"]]
            for asset in assets]



import records


def get_last_saved_alert(db):
    return db.query(
        'SELECT Max(alerts.number) AS max_alert_number FROM alerts'
    )[0]


def get_number_of_local_alerts(db):
    return db.query('SELECT Count(*) as count from alerts')[0].count


def save_new_alert(db, new_alert):
    db.query(
        f'INSERT INTO alerts (id, number) '
        f'VALUES     ("{new_alert.id}", "{new_alert.number}")'
    )


def setup_example_database(db):
    # Create a new example database everytime with some fake data.
    # Tip: In this example we just store the alert number for simplicity
    # but in your application you should store all the alert object
    # metadata.
    # Reference: https://python-matchlightsdk.readthedocs.io/en/latest/api.html#alert
    db.query('DROP TABLE IF EXISTS alerts')
    db.query(
        'CREATE TABLE alerts (id int PRIMARY KEY, number int)'
    )
    db.query(
        'INSERT INTO alerts '
        'VALUES      ("39de2145a1d66330b9d443a84c90d34f", 1)'
    )
    return db



"""Various helper utiliites."""
import calendar
import datetime


__all__ = (
    'blind_email',
    'blind_name',
    'datetime_to_unix',
    'terbium_timestamp_to_datetime',
    'get_matchlight_version',
)


def blind_name(name, width=5):
    """Censors all but the first character of the given string."""
    if name is None:
        name = ''
    return (name[0] if name else '*').ljust(width, '*')


def blind_email(email):
    """Censors an email address."""
    if not email:
        return '*****'
    if '@' in email:
        prefix, suffix = email.split('@', 1)
        suffix = '@' + suffix
    else:
        prefix = email
        suffix = ''

    prefix = prefix[:max(1, min(3, int(len(prefix) / 2)))] + '****'
    return prefix + suffix


def datetime_to_unix(dt):
    """Returns the Unix time for the given datetime object."""
    return calendar.timegm(dt.utctimetuple())


def terbium_timestamp_to_datetime(timestamp):
    """Parses ISO 8601 compatible timestamps.

    Note: This method does not support RFC 3339.

    """
    return datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S')


def get_matchlight_version():
    """Returns Matchlight version."""
    return "2019.7.9.01"



import uuid
import csv


def get_unformatted_data():
    '''Get the plain text document data to fingerprint.

    This is just a placeholder for any function that loads plaintext
    data, in practice this could pull data in from anywhere: a
    database, a CSV, another API, etc.
    '''
    return [{
        'name': 'Internal Record #53',
        'desc': 'Dataset 15 from 2019-01',
        'data': (
            'Ye on properly handsome returned throwing am no whatever. '
            'In without wishing he of picture no exposed talking minutes. '
            'Curiosity continual belonging offending so explained it '
            'exquisite. Do remember to followed yourself material mr '
            'recurred carriage. High drew west we no or at john. About or '
            'given on witty event. Or sociable up material bachelor bringing '
            'landlord confined. Busy so many in hung easy find well up. So '
            'of exquisite my an explained remainder. Dashwood denoting '
            'securing be on perceive my laughing so. Improve him believe '
            'opinion offered met and end cheered forbade. Friendly as '
            'stronger speedily by recurred. Son interest wandered sir '
            'addition end say. Manners beloved affixed picture men ask. '
            'Explain few led parties attacks picture company. On sure fine '
            'kept walk am in it. Resolved to in believed desirous unpacked '
            'weddings together. Nor off for enjoyed cousins herself. '
            'Little our played lively she adieus far sussex. Do theirs '
            'others merely at temper it nearer. '
        )
    }]


def segment_input(input_data, length=840):
    '''Split strings into a list of 840 character strings.'''
    segmented_data = []
    quotient, remainder = divmod(len(input_data), length)
    for i in range(quotient):
        segmented_data += [input_data[i * length:(i + 1) * length]]
    if remainder != 0:
        segmented_data += [input_data[-remainder:]]
    return segmented_data


def format_data(unformatted_data):
    formatted_data = []
    for item in unformatted_data:
        for counter, data_segment in enumerate(segment_input(item['data'])):
            output_line = {}
            output_line['user_record_id'] = uuid.uuid4()
            output_line['data'] = data_segment
            output_line['name'] = f'{item["name"]}-{counter + 1}'
            output_line['desc'] = item['desc']
            formatted_data += [output_line]
    return formatted_data


def save_formatted_data(formatted_data):
    # Tip: You should save this formatted plaintext output so you will
    # be able to connect Matchlight alerts back to source input data.
    with open('formatted-data.csv', 'w', newline='') as csvfile:
        writer = csv.DictWriter(
            csvfile, fieldnames=formatted_data[0].keys()
        )
        writer.writeheader()
        for item in formatted_data:
            writer.writerow(item)


def main():
    unformatted_data = get_unformatted_data()
    print(f'Read in {len(unformatted_data)} unformatted records')
    formatted_data = format_data(unformatted_data)
    save_formatted_data(formatted_data)
    print(f'Saved {len(formatted_data)} formatted records')


if __name__ == '__main__':
    main()



"""Sphinx configuration for the Matchlight SDK."""
# -*- coding: utf-8 -*-
import os
import sys

import setuptools_scm


sys.path.insert(0, os.path.abspath('../../src'))
needs_sphinx = '1.3'
extensions = [
    # 'rst2pdf.pdfbuilder',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = 'Matchlight SDK'
copyright = '2016, Terbium Labs'
author = 'Terbium Labs'
version = setuptools_scm.get_version(root='../..', relative_to=__file__)
release = version
pygments_style = 'sphinx'
html_theme = 'alabaster'
html_theme_options = {
    'fixed_sidebar': True,
    'github_user': 'TerbiumLabs',
    'github_repo': 'python-matchlightsdk',
    'github_button': True,
    'logo': 'logo-stacked-tagline.svg',
    'logo_name': True,
    'sidebar_width': '250px',
}
html_favicon = '_static/favicon.ico'
html_static_path = ['_static']
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
        'donate.html',
    ],
}
# intersphinx_mapping = {'https://docs.python.org/': None}
# pdf_documents = [(
#     'index',
#     'rst2pdf',
#     'Matchlight SDK Documentation',
#     'Terbium Labs',
# )]



"""Unit tests the alert methods of the Matchlight SDK."""
import datetime
import time
import uuid

import responses

import matchlight


@responses.activate
def test_alert_dates(connection, alert, alert_payload):
    """Verifies alert date objects are converted correctly."""
    responses.add(
        method=responses.GET,
        url='{}/alerts?limit=50'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'alerts': [alert_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50)
    assert isinstance(alerts[0].date, datetime.datetime)
    assert isinstance(alerts[0].last_modified, datetime.datetime)
    assert len(responses.calls) == 1


@responses.activate
def test_alert_filter(connection, alert, alert_payload):
    """Verifies alert listing and filtering."""
    responses.add(
        method=responses.GET,
        url='{}/alerts?limit=50'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'alerts': [alert_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50)
    assert len(alerts) == 1
    assert alerts[0].id == alert.id
    assert len(responses.calls) == 1


@responses.activate
def test_alert_filter_seen(connection, alert, alert_payload):
    """Verifies alert filtering on 'seen'."""
    # Create opposite alert
    unseen_payload = alert_payload.copy()
    unseen_payload['seen'] = 'false'
    unseen_payload['id'] = str(uuid.uuid4())

    # Get seen alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?seen=1&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2
        ),
        json={'alerts': [alert_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50, seen=True)
    assert len(alerts) == 1
    assert alerts[0].id == alert_payload['id']
    assert len(responses.calls) == 1

    # Get unseen alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?seen=0&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2
        ),
        json={'alerts': [unseen_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50, seen=False)
    assert len(alerts) == 1
    assert alerts[0].id == unseen_payload['id']
    assert len(responses.calls) == 2


@responses.activate
def test_alert_filter_archived(connection, alert, alert_payload):
    """Verifies alert filtering on 'archived'."""
    # Create opposite alert
    unarchived_payload = alert_payload.copy()
    unarchived_payload['archived'] = 'false'
    unarchived_payload['id'] = str(uuid.uuid4())

    # Get archived alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?archived=1&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2
        ),
        json={'alerts': [alert_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50, archived=True)
    assert len(alerts) == 1
    assert alerts[0].id == alert_payload['id']
    assert len(responses.calls) == 1

    # Get unarchived alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?archived=0&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2
        ),
        json={'alerts': [unarchived_payload]},
        status=200
    )
    alerts = connection.alerts.filter(limit=50, archived=False)
    assert len(alerts) == 1
    assert alerts[0].id == unarchived_payload['id']
    assert len(responses.calls) == 2


@responses.activate
def test_alert_filter_project(connection, alert, alert_payload, project):
    """Verifies alert filtering on 'upload_token'."""
    responses.add(
        method=responses.GET,
        url='{}/alerts?upload_token_filter={}&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        json={'alerts': [alert_payload]},
        status=200
    )

    alerts = connection.alerts.filter(limit=50, project=project)
    assert len(alerts) == 1
    assert alerts[0].id == alert_payload['id']
    assert len(responses.calls) == 1


@responses.activate
def test_alert_filter_record(connection, alert, alert_payload,
                             document_record):
    """Verifies alert filtering on 'record_id'."""
    responses.add(
        method=responses.GET,
        url='{}/alerts?record_id_filter={}&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            document_record.id
        ),
        json={'alerts': [alert_payload]},
        status=200
    )

    alerts = connection.alerts.filter(limit=50, record=document_record)
    assert len(alerts) == 1
    assert alerts[0].id == alert_payload['id']
    assert len(responses.calls) == 1


@responses.activate
def test_alert_filter_mtime(connection, alert, alert_payload):
    """Verifies alert filtering on 'seen'."""
    now = int(time.time())
    # Two days ago
    old_payload = alert_payload.copy()
    old_payload['mtime'] = now - 172800
    old_payload['id'] = str(uuid.uuid4())

    # Get all alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?mtime={}&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            now - 259200
        ),
        json={'alerts': [alert_payload, old_payload]},
        status=200
    )

    alerts = connection.alerts.filter(
        limit=50,
        last_modified=datetime.datetime.utcfromtimestamp(now - 259200)
    )
    assert len(alerts) == 2
    assert len(responses.calls) == 1

    # Get new alerts
    responses.add(
        method=responses.GET,
        url='{}/alerts?mtime={}&limit=50'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            now - 86400
        ),
        json={'alerts': [alert_payload]},
        status=200
    )

    alerts = connection.alerts.filter(
        limit=50,
        last_modified=datetime.datetime.utcfromtimestamp(now - 86400)
    )
    assert len(alerts) == 1
    assert alerts[0].id == alert_payload['id']
    assert len(responses.calls) == 2


@responses.activate
def test_alert_edit(connection, alert, alert_payload):
    """Verifies alert editing."""
    # Do nothing
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=rsps.POST,
            url='{}/alert/{}/edit'.format(
                matchlight.MATCHLIGHT_API_URL_V2,
                alert.id,
            ),
            json={
                'archived': True,
                'seen': True
            },
            status=200
        )

        response = connection.alerts.edit(alert.id)
        assert response['seen'] is True
        assert response['archived'] is True

    # Un-archive
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=rsps.POST,
            url='{}/alert/{}/edit'.format(
                matchlight.MATCHLIGHT_API_URL_V2,
                alert.id,
            ),
            json={
                'archived': False,
                'seen': True
            },
            status=200
        )
        response = connection.alerts.edit(alert.id, archived=False)

        assert response['seen'] is True
        assert response['archived'] is False

    # Un-see
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=rsps.POST,
            url='{}/alert/{}/edit'.format(
                matchlight.MATCHLIGHT_API_URL_V2,
                alert.id,
            ),
            json={
                'archived': False,
                'seen': False
            },
            status=200
        )
        response = connection.alerts.edit(alert.id, seen=False)
        assert response['seen'] is False
        assert response['archived'] is False

    # Both
    with responses.RequestsMock() as rsps:
        rsps.add(
            method=rsps.POST,
            url='{}/alert/{}/edit'.format(
                matchlight.MATCHLIGHT_API_URL_V2,
                alert.id,
            ),
            json={
                'archived': True,
                'seen': True
            },
            status=200
        )
        response = connection.alerts.edit(alert.id, seen=True, archived=True)
        assert response['seen'] is True
        assert response['archived'] is True


@responses.activate
def test_alert_details(connection, alert, alert_details_pii_payload):
    """Verifies alert get details responses."""
    responses.add(
        method=responses.GET,
        url='{}/alert/{}/details'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            alert.id
        ),
        json=alert_details_pii_payload,
        status=200
    )

    details_ = connection.alerts.get_details(alert)
    assert details_ == alert_details_pii_payload
    assert len(responses.calls) == 1

    details_ = connection.alerts.get_details(alert.id)
    assert details_ == alert_details_pii_payload
    assert len(responses.calls) == 2



"""An interface for searching the Matchlight fingerprints database."""
from __future__ import absolute_import

import datetime
import json

import six

import matchlight.error
import pylibfp


class SearchMethods(object):
    """Provides methods for interfacing with the search API."""

    def __init__(self, ml_connection):  # noqa: D205,D400
        """Initializes a search interface with the given Matchlight
        connection.

        Args:
            ml_connection (:class:`~.Connection`): A Matchlight
                connection instance.

        """
        self.conn = ml_connection

    def search(self, query=None, email=None, ssn=None, phone=None,
               fingerprints=None):
        """Performs a Matchlight search.

        Provides a retrospective search capability. User can only
        perform one search type at time. Search type is specified using
        keyword arguments.

        Example:

            Search for text::

                >>> ml.search(query="magic madness heaven sin")

            Search for an email address::

                >>> ml.search(email="familybird@terbiumlabs.com")

            Search for a social security number::

                >>> ml.search(ssn="000-00-0000")

            Search for a phone number::

                >>> ml.search(phone="804-222-1111")

        Args:
            query (:obj:`str`, optional): A text query.
            email (:obj:`str`, optional): A valid email address.
            ssn (:obj:`str`, optional): A social security number.
            phone (:obj:`str`, optional): A phone number.
            fingerprints (:obj:`list` of :obj:`str`, optional): A sequence of
            Matchlight fingerprints, these will be searched as if one query.


        Returns:
            :obj:`list` of :obj:`dict`: Each search result returns a
                score, url, ts.

        """
        # only search for one thing at a time.
        if sum([1 for k in [query, fingerprints, email, ssn, phone]
                if k is not None]) != 1:
            raise matchlight.error.SDKError(
                'Input Error: Must specify exactly one search type per call.')

        if email:
            fingerprints = pylibfp.fingerprints_pii_email_address(str(email))
        elif phone:
            fingerprints = pylibfp.fingerprints_pii_phone_number(str(phone))
        elif ssn:
            fingerprints = pylibfp.fingerprints_pii_ssn(str(ssn))
        elif query:
            result_json = pylibfp.fingerprint(
                query, flags=pylibfp.OPTIONS_TILED)
            result = json.loads(result_json)
            fingerprints = result['data']['fingerprints']

        # We have to convert possible lists of lists to a flat list of strings
        def flatten_iter(x):
            if not isinstance(x, list):
                yield(x)
            else:
                for i in x:
                    # same as yield from flatten_iter(i)
                    for j in flatten_iter(i):
                        yield j

        data = {'fingerprints': list(flatten_iter(fingerprints))}
        response = self.conn.request(
            '/detailed_search',
            data=json.dumps(data),
            endpoint=self.conn.search_endpoint,
        )
        try:
            results = response.json()['results']
        except KeyError:
            raise matchlight.error.SDKError('Failed to get search results')
        for result in results:
            for url in result['urls']:
                yield {
                    # PII Alerts are always 800, Search results on PII filds
                    # should be as well.
                    'score': 800 if any(
                        [email, ssn, phone]
                    ) else result['score'],
                    'ts': datetime.datetime.fromtimestamp(float(url[0])),
                    'url': url[1]
                }

    def pii_search(self, email=None, limit=50):
        """Performs a Matchlight search specifically for PII.

        Provides a retrospective search capability designed specifically for
        finding compromised PII data.
        Search results are sorted & show which fields matched on each hit.
        Only exact matches are returned.

        Example:

            >>> ml.pii_search(email="familybird@terbiumlabs.com")

        Args:
            email (:obj:`str`, required): A valid email address.
            limit (:obj:`int`, optional): The number of Alerts to return,
                defaults to 50.

        Returns:
            :obj:`list` of :obj:`dict`: Each search result returns a
                source, ts, fields

        """
        if not any(email):
            raise matchlight.error.SDKError(
                'Input Error: At least one field is required.'
            )

        request_data = {}

        if email:
            request_data['email_fingerprints'] = (
                pylibfp.fingerprints_pii_email_address(str(email))
            )

        if limit:
            request_data['limit'] = limit

        response = self.conn.request(
            '/pii_search',
            data=json.dumps(request_data),
            endpoint=self.conn.search_endpoint,
        )
        try:
            results = response.json()['results']
        except KeyError:
            raise matchlight.error.SDKError('Failed to get search results')
        for result in results:
            # This result can seemingly be in different formats.
            try:
                result['ts'] = int(result['ts'])
            except ValueError:
                pass
            if isinstance(result['ts'], six.text_type):
                result['ts'] = datetime.datetime.strptime(
                    result['ts'],
                    '%Y-%m-%dT%H:%M:%S'
                )
            elif isinstance(result['ts'], int):
                result['ts'] = datetime.datetime.fromtimestamp(
                    float(result['ts'])
                )

            yield result



import csv
import json

from matchlight import Matchlight


def get_document_data():
    output = []
    with open('formatted-data.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            output += [row]
    return output


def generate_fingerprints(document_data):
    '''Generate fingerprints but do not upload them to Matchlight.'''
    # Tip: Your access and secret keys can be blank here because we're
    # not uploading anything yet.
    ml = Matchlight(access_key='', secret_key='')

    fingerprinted_data = []
    for counter, item in enumerate(document_data):
        fingerprint_json_blob = ml.records.add_document(
            project='',
            name=item['name'],
            description=item['desc'],
            content=item['data'],
            user_record_id=item['user_record_id'],
            offline=True
        )
        fingerprinted_data += [fingerprint_json_blob]
    return fingerprinted_data


def save_fingerprints(fingerprinted_data):
    '''Save a local copy of the fingerprinted data.

    Using JSON instead of CSV because the fingerprinted data is nested
    as opposed to flat.
    '''
    with open('fingerprints.json', 'w', newline='') as jsonfile:
        json.dump(fingerprinted_data, jsonfile)


def main():
    '''Create an offline set of Matchlight document fingerprints.'''
    document_data = get_document_data()
    fingerprinted_data = generate_fingerprints(document_data)
    save_fingerprints(fingerprinted_data)
    print(f'Fingerprinted {len(fingerprinted_data)} records')


if __name__ == '__main__':
    main()



import csv
import json

from matchlight import Matchlight


def load_fingerprints():
    '''Load data generated by fingerprint-documents.py'''
    output = []
    with open('fingerprints.json', 'r') as jsonfile:
        return json.load(jsonfile)


def main():
    '''Upload a previsouly fingerprinted set of documents.'''
    fingerprint_data = load_fingerprints()

    # Tip: You can generate your Matchlight API keys through the
    # Matchlight web interface
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html?highlight=keys#authentication
    ml = Matchlight(
        access_key='your-matchlight-api-access-key',
        secret_key='your-matchlight-api-secret-key'
    )
    # Tip: Make sure you have already created a project
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html#create-a-new-project
    project = ml.projects.get('your-project-upload-token')

    total_records = len(fingerprint_data)
    for counter, document_fingerprint in enumerate(fingerprint_data):
        print(
            f'Uploading {document_fingerprint["name"]} '
            f'({counter} of {total_records})'
        )
        ml.records.add_document_from_fingerprints(
            project, document_fingerprint
        )


if __name__ == '__main__':
    main()



"""Unit tests the project methods of the Matchlight SDK."""
import datetime
import json

import pytest
import responses
import six.moves

import matchlight


PROJECT_TYPES = ('document', 'pii', 'source_code')


def test_project_last_modified(project):
    """Verifies that a project's ``last_date_modified`` is parsed."""
    assert project.last_modified == datetime.datetime.fromtimestamp(
        project.last_date_modified)


@responses.activate
def test_project_add(connection, project_name, project_payload,
                     project_type, upload_token):
    """Verifies project creation."""
    responses.add(
        method=responses.POST,
        url='{}/project/add'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'data': project_payload},
        status=200
    )
    project = connection.projects.add(project_name, project_type)
    assert project.upload_token == upload_token
    assert len(responses.calls) == 1


@responses.activate
def test_project_edit(connection, project, project_payload):
    """Verifies project renaming."""
    responses.add(
        method=responses.POST,
        url='{}/project/{}/edit'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token,
        ),
        json='{}',
        status=200
    )
    new_name = 'Test Project 1'
    connection.projects.edit(project, new_name)
    assert project.name == new_name
    assert len(responses.calls) == 1

    new_name = 'Test Project 2'
    responses.add(
        method=responses.GET,
        url='{}/project/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token,
        ),
        json=project_payload,
        status=200
    )
    project = connection.projects.edit(project, new_name)
    assert project.name == new_name
    assert len(responses.calls) == 2


@responses.activate
def test_project_delete(connection, project, project_payload):
    """Verifies project deletion."""
    responses.add(
        method=responses.POST,
        url='{}/project/{}/delete'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token,
        ),
        json='{}',
        status=200
    )
    # responses.add(
    #     method=responses.GET,
    #     url='{}/project/{}'.format(
    #         matchlight.MATCHLIGHT_API_URL_V2,
    #         project.upload_token,
    #     ),
    #     json=project_payload,
    #     status=200
    # )
    connection.projects.delete(project)
    assert len(responses.calls) == 1


@responses.activate
def test_project_filter(connection, project_payload, project):
    """Verifies project listing and filtering by type."""
    responses.add(
        method=responses.GET,
        url='{}/projects'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'data': [project_payload]},
        status=200
    )
    projects = connection.projects.filter()
    assert len(projects) == 1
    assert projects[0].upload_token == project.upload_token
    assert len(responses.calls) == 1

    project_list = [project_payload]
    for _ in six.moves.range(5):
        payload = project_payload.copy()
        for project_type in PROJECT_TYPES:
            if project_type == payload['project_type']:
                continue
            payload['project_type'] = project_type
            break
        project_list.append(payload)
    responses.add(
        method=responses.GET,
        url='{}/projects'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'data': project_list},
        status=200
    )
    projects = connection.projects.filter(project_type=project.project_type)
    assert len(projects) == 1
    assert projects[0].project_type == project.project_type
    assert len(responses.calls) == 2


@responses.activate
def test_project_get(connection, project_payload, project):
    """Verifies project retrieval."""
    responses.add(
        method=responses.GET,
        url='{}/project/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        body=json.dumps(project_payload),
        status=200
    )
    responses.add(
        method=responses.GET,
        url='{}/project/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        body='{}',
        status=404
    )
    responses.add(
        method=responses.GET,
        url='{}/project/{}'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            project.upload_token
        ),
        body='{}',
        status=500
    )

    project_ = connection.projects.get(project.upload_token)
    assert project.upload_token == project_.upload_token
    assert len(responses.calls) == 1

    project_ = connection.projects.get(project.upload_token)
    assert project_ is None
    assert len(responses.calls) == 2

    with pytest.raises(matchlight.error.APIError):
        connection.projects.get(project.upload_token)
    assert len(responses.calls) == 3


@responses.activate
def test_project_iteration(connection, project, project_payload):
    """Verifies project iteration."""
    responses.add(
        method=responses.GET,
        url='{}/projects'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'data': [project_payload]},
        status=200,
    )
    projects_iterable = iter(connection.projects)
    assert next(projects_iterable).details == project.details
    with pytest.raises(StopIteration):
        next(projects_iterable)
    assert len(responses.calls) == 1



import records
from matchlight import Matchlight

from local_alert_utilities import (
    get_last_saved_alert,
    get_number_of_local_alerts,
    save_new_alert,
    setup_example_database,
)


def get_and_store_alerts():
    # Create a database connection and set up a local alert database.
    # db = records.Database('sqlite:///matchlight-alerts-example.db')
    db = records.Database('sqlite:///:memory:')
    setup_example_database(db)
    print(
        f'Number of alerts in database is {get_number_of_local_alerts(db)}'
    )

    # Create a Matchliht API connection
    # Tip: You can generate your Matchlight API keys through the
    # Matchlight web interface
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html?highlight=keys#authentication
    ml = Matchlight(
        access_key='your-matchlight-api-access-key',
        secret_key='your-matchlight-api-secret-key'
    )

    # Tip: Make sure you have already created a project
    # https://python-matchlightsdk.readthedocs.io/en/latest/guide.html#create-a-new-project
    project = ml.projects.get('your-project-upload-token')

    # Find the last saved alert in our database. We will work backwards
    # through the new alerts until we get to this value.
    last_saved_alert = get_last_saved_alert(db)
    print(f'Latest alert is number {last_saved_alert.max_alert_number}')

    keep_going = True
    last_alert_query_parameter = None
    while keep_going:
        print(f'Last Alert Query Parameter is {last_alert_query_parameter}')
        alert_query = ml.alerts.filter(
            project=project, limit=2, last_alert=last_alert_query_parameter
        )
        if len(alert_query) == 0:
            print('No alerts left')
            break
        else:
            new_alert = alert_query[0]
        print(f'New Alert found with Alert Number {new_alert.number}')
        if new_alert.number > last_saved_alert.max_alert_number:
            save_new_alert(db, new_alert)
            last_alert_query_parameter = new_alert.number
        else:
            keep_going = False
            print('No new alerts remain')
    print(
        f'Number of alerts in databasse is {get_number_of_local_alerts(db)}'
    )


if __name__ == '__main__':
    get_and_store_alerts()



#!/usr/bin/env python
"""Setup script for Matchlight SDK."""
import io
import os
import sys

import setuptools


def setup():  # noqa: D103
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    setup_requirements = ['setuptools_scm']
    if needs_sphinx:
        setup_requirements.append('sphinx>=1.3')
    if needs_pytest:
        setup_requirements.append('pytest-runner')
    install_requirements = ['requests[security]', 'six']
    test_requirements = [
        'httpretty',
        'mock',
        'pytest>=2.8.0',
        'pytest-cov',
        'pytest-httpretty',
    ]
    # Readthedocs requires Sphinx extensions to be specified as part of
    # install_requires in order to build properly.
    on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
    if on_rtd:
        install_requirements.extend(setup_requirements)
    if sys.version_info < (3,):
        install_requirements.append('backports.csv')
    with io.open('README.rst') as fp:
        readme = fp.read()
    setuptools.setup(
        name='matchlightsdk',
        license='BSD',
        description=('Software development kit for the Terbium Labs '
                     'Matchlight product.'),
        long_description=readme + '\n',
        author='Terbium Labs',
        author_email='developers@terbiumlabs.com',
        url='https://terbiumlabs.com',
        packages=setuptools.find_packages('src'),
        package_dir={'': 'src'},
        # use_scm_version=True,
        version='2019.7.9.01',
        package_data={
            '': ['LICENSE', 'LICENSES'],
            'pylibfp': [
                'lib/*.so',
                'lib/*.dll',
                'lib/*.dylib',
            ],
        },
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Intended Audience :: Financial and Insurance Industry',
            'Intended Audience :: Healthcare Industry',
            'Intended Audience :: Information Technology',
            'Intended Audience :: Legal Industry',
            'Intended Audience :: Science/Research',
            'Natural Language :: English',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        install_requires=install_requirements,
        setup_requires=setup_requirements,
        tests_require=test_requirements,
    )


if __name__ == '__main__':
    setup()



"""An interface for creating and retrieving PII records in Matchlight."""
from __future__ import absolute_import

import io
import json

import six

import matchlight.error
import matchlight.utils

from pylibfp import (
    fingerprint,
    fingerprints_pii_address_variants,
    fingerprints_pii_city_state_zip_variants,
    fingerprints_pii_credit_card,
    fingerprints_pii_email_address,
    fingerprints_pii_iban,
    fingerprints_pii_medicare_id,
    fingerprints_pii_name_variants,
    fingerprints_pii_passport,
    fingerprints_pii_phone_number,
    fingerprints_pii_ssn,
    MODE_CODE,
    OPTIONS_TILED,
)


__all__ = (
    'Record',
    'RecordMethods',
)


MAX_DOCUMENT_FINGERPRINTS = 840


class Record(object):
    """Represents a personal information record."""

    def __init__(self, id, name, description, ctime=None, mtime=None,
                 metadata=None):
        """Initializes a new personal information record.

        Args:
            id (:obj:`str`): A 128-bit UUID.
            name (:obj:`str`): The name of the record.
            description (:obj:`str`): The description of the record.
            ctime (:obj:`int`, optional): A Unix timestamp of the
                record creation timestamp.
            mtime (:obj:`int`, optional): A Unix timestamp of the
                record last modification date timestamp.

        """
        if metadata is None:
            metadata = {}
        self.id = id
        self.name = name
        self.description = description
        self.ctime = ctime
        self.mtime = mtime
        self.metadata = metadata

    @classmethod
    def from_mapping(cls, mapping):
        """Creates a new project instance from the given mapping."""
        return cls(
            id=mapping['id'],
            name=mapping['name'],
            description=mapping['description'],
            ctime=mapping['ctime'],
            mtime=mapping['mtime'],
            metadata=mapping['metadata'],
        )

    @property
    def user_provided_id(self):
        """:obj:`int`: The user provided record identifier."""
        return self.metadata.get('user_record_id', None)

    @property
    def details(self):
        """:obj:`dict`: Returns the feed details as a mapping."""
        return {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'ctime': self.ctime,
            'mtime': self.mtime,
            'metadata': self.metadata,
        }

    def __repr__(self):  # pragma: no cover
        return '<Record(name="{}", id="{}")>'.format(self.name, self.id)


class RecordMethods(object):
    """Provides methods for interfacing with the records API.

    Examples:
        Get record by record id::
            >>> record = ml.records.get("0760570a2c4a4ea68d526f58bab46cbd")
            >>> record
            <Record(name="pce****@terbiumlabs.com",
            id="0760570a2c4a4ea68d526f58bab46cbd")>

        Add PII records to a project::
            >>> pii_project = ml.projects.add(
            ...     name="Employee Database May 2016",
            ...     project_type="pii")
            >>> record_data = {
            ...     "first_name": "Bird",
            ...     "last_name": "Feather",
            ...     "email": "familybird@teribumlabs.com",
            ... }
            >>> new_record = ml.records.add_pii(
            ...      pii_project,
            ...      "uploaded on 20160519",
            ...      **record_data)

        Delete a record::
            >>> record
            <Record(name="fam****@terbiumlabs.com",
            id="655a732ad0f243beab1801651c2088a3")>
            >>> ml.record.delete(record)

    """

    def __init__(self, ml_connection):  # noqa: D205,D400
        """Initializes a records interface with the given Matchlight
        connection.

        Args:
            ml_connection (:class:`~.Connection`): A Matchlight
                connection instance.

        """
        self.conn = ml_connection

    def all(self):
        """Returns all records associated with the account."""
        return self.filter()

    def add_document(self, project, name, description, content,
                     user_record_id='-', min_score=None, offline=False):
        """Creates a new document record in the given project.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            name (:obj:`str`): The name of the document (not
                fingerprinted).
            description (:obj:`str`): A description of the record (not
                fingerprinted).
            content (:obj:`str`): The text of the document to be
                fingerprinted. Must be 840 characters or less.
            user_record_id (:obj:`str`, optional): An optional, user
                provided custom record identifier. Defaults to
                :obj:`NoneType`.
            offline (:obj:`bool`, optional): Run in "offline mode". No
                data is sent to the Matchlight server. Returns a
                dictionary of values instead of a :class:`~.Report`
                instance.

        Returns:
            :class:`~.Record`: Created record with metadata.

        """
        if len(content) > MAX_DOCUMENT_FINGERPRINTS:
            raise matchlight.error.SDKError(
                'Fingerprinter Failed: the maximum length of a Document record'
                ' is 840 characters.'
            )

        result_json = fingerprint(content, flags=OPTIONS_TILED)
        result = json.loads(result_json)
        fingerprints = result['data']['fingerprints']

        data = {
            'name': name,
            'desc': description,
            'user_record_id': user_record_id,
            'fingerprints': fingerprints,
            'metadata': {
                'fingerprinting_tool_name': 'Python SDK',
                'fingerprinting_tool_version': matchlight.__version__
            }
        }
        if min_score is not None:
            data['metadata']['min_score'] = str(min_score)
        if offline:
            return data
        else:
            return self.add_document_from_fingerprints(project, data)

    def add_document_from_fingerprints(self, project, fingerprint_data):
        """Add a document record from fingerprints.

        Add a document record from fingerprinted data generated by the
        :class:`~/.Record.add_pii` in offline mode.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            fingerprint_data (:obj:`dict`): The output of
                :class:`~/.Record.add_document(offline=True)`
        """
        response = self.conn.request(
            '/records/upload/document/{upload_token}'.format(
                upload_token=project.upload_token
            ),
            data=json.dumps(fingerprint_data)
        )
        return Record.from_mapping(response.json())

    def add_pii(self, project, description, email, first_name=None,
                middle_name=None, last_name=None, ssn=None, address=None,
                city=None, state=None, zipcode=None, phone=None,
                credit_card=None, medicare_id=None, passport=None, iban=None,
                user_record_id='-', offline=False):
        """Creates a new PII record in the given project.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            description (:obj:`str`): A description of the record (not
                fingerprinted).
            email (:obj:`str`, optional): An email address.
            first_name (:obj:`str`, optional): Defaults to
                :obj:`NoneType`.
            middle_name (:obj:`str`, optional): Defaults to
                :obj:`NoneType`.
            last_name (:obj:`str`, optional): Defaults to
                :obj:`NoneType`.
            ssn (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            address (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            city (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            state (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            zipcode (int, optional): Defaults to :obj:`NoneType`.
            phone (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            credit_card (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            medicare_id (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            passport (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            iban (:obj:`str`, optional): Defaults to :obj:`NoneType`.
            user_record_id (:obj:`str`, optional): An optional, user
                provided custom record identifier. Defaults to
                :obj:`NoneType`.
            offline (:obj:`bool`, optional): Run in "offline mode". No
                data is sent to the Matchlight server. Returns a
                dictionary of values instead of a :class:`~.Report`
                instance.

        Returns:
            :class:`~.Record`: Created record with metadata.

        """
        if first_name is not None and last_name is None:
            raise matchlight.error.SDKError(
                'Fingerprinter Failed: the last_name argument is required '
                'along with the first_name argument.'
            )

        if first_name is None and last_name is not None:
            raise matchlight.error.SDKError(
                'Fingerprinter Failed: the first_name argument is required '
                'along with the last_name argument.'
            )

        data = {
            'desc': description,
            'user_record_id': user_record_id,
            'blinded_first': matchlight.utils.blind_name(first_name),
            'blinded_last': matchlight.utils.blind_name(last_name),
            'blinded_email': matchlight.utils.blind_email(email),
            'metadata': {
                'fingerprinting_tool_name': 'Python SDK',
                'fingerprinting_tool_version': matchlight.__version__
            }
        }

        if any((first_name, middle_name, last_name)):
            name_fingerprints = fingerprints_pii_name_variants(
                first_name or '', middle_name or None, last_name or '')
            data['name_fingerprints'] = name_fingerprints

        if email:
            email_fingerprints = fingerprints_pii_email_address(email)
            data['email_fingerprints'] = email_fingerprints
        data['blinded_email'] = matchlight.utils.blind_email(email)
        data['name'] = matchlight.utils.blind_email(email)

        if ssn:
            ssn_fingerprints = [fingerprints_pii_ssn(ssn)]
            data['ssn_fingerprints'] = ssn_fingerprints

        if address:
            address_fingerprints = fingerprints_pii_address_variants(
                address)
            data['street_address_fingerprints'] = address_fingerprints

        if any((city, state, zipcode)):
            csz_fingerprints = fingerprints_pii_city_state_zip_variants(
                *[six.text_type(text) if text is not None else ''
                  for text in (city, state, zipcode)])
            data['city_state_zip_fingerprints'] = csz_fingerprints

        if phone:
            phone_fingerprints = fingerprints_pii_phone_number(phone)
            data['phone_fingerprints'] = [phone_fingerprints]

        if credit_card:
            cc_fingerprints = fingerprints_pii_credit_card(credit_card)
            data['credit_card_fingerprints'] = [cc_fingerprints]

        if medicare_id:
            medicare_id_fingerprints = fingerprints_pii_medicare_id(
                medicare_id
            )
            data['medicare_id_fingerprints'] = [medicare_id_fingerprints]

        if passport:
            passport_fingerprints = fingerprints_pii_passport(passport)
            data['passport_fingerprints'] = [passport_fingerprints]

        if iban:
            iban_fingerprints = fingerprints_pii_iban(iban)
            data['iban_fingerprints'] = iban_fingerprints

        if offline:
            return data
        else:
            return self.add_pii_from_fingerprints(project, data)

    def add_pii_from_fingerprints(self, project, fingerprint_data):
        """Add a PII record from fingerprints.

        Add a PII record from fingerprinted data generated by the
        :class:`~/.Record.add_pii` in offline mode.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            fingerprint_data (:obj:`dict`): The output of
                :class:`~/.Record.add_pii(offline=True)`
        """
        response = self.conn.request(
            '/records/upload/pii/{}'.format(
                project.upload_token
            ),
            data=json.dumps(fingerprint_data)
        )
        return Record.from_mapping(response.json())

    def add_source_code(self, project, name, description, code_path,
                        min_score=None, offline=False):
        """Creates a new source code record in the given project.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            name (:obj:`str`): The name of the file (not
                fingerprinted).
            description (:obj:`str`): A description of the code (not
                fingerprinted).
            code_path (:obj:`str`): The location of the source code.
                Code must be 840 characters or less.
            user_record_id (:obj:`str`, optional): An optional, user
                provided custom record identifier. Defaults to
                :obj:`NoneType`.
            offline (:obj:`bool`, optional): Run in "offline mode". No
                data is sent to the Matchlight server. Returns a
                dictionary of values instead of a :class:`~.Report`
                instance.

        Returns:
            :class:`~.Record`: Created record with metadata.

        """
        with io.open(code_path, 'r', encoding='utf-8') as document:
            content = document.read()

        if len(content) > MAX_DOCUMENT_FINGERPRINTS:
            raise matchlight.error.SDKError(
                'Fingerprinter Failed: the maximum length of a Source Code '
                'record is 840 characters.'
            )

        result_json = fingerprint(content, flags=OPTIONS_TILED, mode=MODE_CODE)
        result = json.loads(result_json)
        fingerprints = result['data']['fingerprints']

        data = {
            'name': name,
            'desc': description,
            'fingerprints': fingerprints,
            'metadata': {
                'fingerprinting_tool_name': 'Python SDK',
                'fingerprinting_tool_version': matchlight.__version__
            }
        }

        if min_score is not None:
            data['metadata']['min_score'] = str(min_score)
        if offline:
            return data
        else:
            return self.add_source_code_from_fingerprints(project, data)

    def add_source_code_from_fingerprints(self, project, fingerprint_data):
        """Add a source code record from fingerprints.

        Add a souce code record from fingerprinted data generated by the
        :class:`~/.Record.add_source_code` in offline mode.

        Args:
            project (:class:`~.Project`): Project object to associate
                with record.
            fingerprint_data (:obj:`dict`): The output of
                :class:`~/.Record.add_source_code(offline=True)`
        """
        response = self.conn.request(
            '/records/upload/source_code/{}'.format(
                upload_token=project.upload_token
            ),
            data=json.dumps(fingerprint_data)
        )
        return Record.from_mapping(response.json())

    def delete(self, record_or_id):
        """Delete a fingerprinted record.

        Args:
            record_or_id (:class:`~.Record` or :obj:`str`): The record
                object or identifier to be deleted.

        Returns:
            :obj:`NoneType`

        """
        if isinstance(record_or_id, Record):
            record_upload_token = record_or_id.id
        else:
            record_upload_token = record_or_id
        self.conn.request('/record/{}/delete'.format(record_upload_token),
                          data=json.dumps({}))

    def filter(self, project=None):
        """Returns a list of records.

        Providing an optional **project** keyword argument will only
        return records that are associated with a specific project.

        Example:
            Request all records::

                >>> my_project
                <Project(name="Super Secret Algorithm", type="source_code")>
                >>> ml.records.filter(project=my_project)
                [<Record(name="fam****@fakeemail.com",
                id="625a732ad0f247beab18595z951c2088a3")>,
                Record(name="pce****@fakeemail.com",
                id="f9427dd5a24d4a98b2069004g04c2977")]

        Args:
            project (:class:`~.Project`, optional): a project object.
                Defaults to all projects if not specified.

        Returns:
            :obj:`list` of :class:`~.Record`: List of records that
                are associated with a project.

        """
        if project is not None:
            upload_token = project.upload_token
        else:
            upload_token = None
        response = self.conn.request('/records', params={
            'upload_token': upload_token})
        records = []
        for payload in response.json().get('data', []):
            records.append(Record(
                id=payload['id'],
                name=payload['name'],
                description=payload['description'],
                ctime=int(payload['ctime']),
                mtime=int(payload['mtime']),
            ))
        return records

    def get(self, record_id):
        """Returns a record by the given record ID.

        Args:
            record_id (:obj:`str`): The record identifier.

        Returns:
           :class:`~.Record`: A record instance.

        """
        return next((record for record in self.filter()
                     if record.id == record_id), None)

    def __iter__(self):
        return iter(self.filter())



"""Unit tests the alert methods of the Matchlight SDK."""
import datetime
import json

import responses

import matchlight


@responses.activate
def test_pii_search(connection, pii_search_email_only_results):
    """Verifies alert listing and filtering."""
    responses.add(
        method=responses.POST,
        url='{}/pii_search'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'results': pii_search_email_only_results},
        status=200
    )

    assert list(
        connection.pii_search(email='familybird@terbiumlabs.com')
    ) == [
        {
            'fields': ['email'],
            'ts': datetime.datetime.strptime(
                '2018-07-25T20:00:44',
                '%Y-%m-%dT%H:%M:%S'
            ),
            'source': 'Exactis Breach June 2018'
        },
        {
            'fields': ['email'],
            'ts': datetime.datetime.strptime(
                '2017-01-25T02:35:04',
                '%Y-%m-%dT%H:%M:%S'
            ),
            'source': 'https://pastebin.com/raw.php?i=1DgbtSZc'
        },
        {
            'fields': ['email'],
            'ts': datetime.datetime.fromtimestamp(1556221970.0),
            'source': 'Zoosk Breach Nov 2016'
        },
        {
            'fields': ['email'],
            'ts': datetime.datetime.fromtimestamp(1558333205.0),
            'source': 'https://www.reddit.com/r/AskReddit/comments/3oqj4a'
        }
    ]

    assert json.loads(responses.calls[0].request.body) == {
        'email_fingerprints': [['ff71225ace46c2b0'], ['293de73a18a5e063']],
        'limit': 50,
    }


@responses.activate
def test_pii_search_limit(connection, pii_search_email_only_results):
    """Verifies alert listing and filtering."""
    responses.add(
        method=responses.POST,
        url='{}/pii_search'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'results': pii_search_email_only_results[:2]},
        status=200
    )

    assert list(
        connection.pii_search(email='familybird@terbiumlabs.com', limit=2)
    ) == [
        {
            'fields': ['email'],
            'ts': datetime.datetime.strptime(
                '2018-07-25T20:00:44',
                '%Y-%m-%dT%H:%M:%S'
            ),
            'source': 'Exactis Breach June 2018'
        },
        {
            'fields': ['email'],
            'ts': datetime.datetime.strptime(
                '2017-01-25T02:35:04',
                '%Y-%m-%dT%H:%M:%S'
            ),
            'source': 'https://pastebin.com/raw.php?i=1DgbtSZc'
        }
    ]

    assert json.loads(responses.calls[0].request.body) == {
        'email_fingerprints': [['ff71225ace46c2b0'], ['293de73a18a5e063']],
        'limit': 2,
    }


@responses.activate
def test_search_email(connection, search_email_only_results):
    """Verifies generic search when using email."""
    responses.add(
        method=responses.POST,
        url='{}/detailed_search'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'results': search_email_only_results},
        status=200
    )

    assert list(
        connection.search(email='familybird@terbiumlabs.com')
    ) == [
        {
            'score': 800,
            'ts': datetime.datetime.fromtimestamp(1453298293.0),
            'url': (
                'http://blockchainbdgpzk.onion/tx/4f4097992b89156'
                '690817556fc3f540535bdfadde06661c9cae21d500943f970'
            )
        }, {
            'score': 800,
            'ts': datetime.datetime.fromtimestamp(1438870917.0),
            'url': (
                'http://nqigfqrxnkwcqmiq.onion/'
                'wiki/index.php#Whistleblowing'
            )
        }
    ]



"""Unit tests the feed methods of the Matchlight SDK."""
import csv
import datetime
import io
import json
import random
import time
import uuid

import mock
import pytest
import responses
import six.moves

import matchlight


@pytest.fixture(scope='function')
def start_time():
    """A feed start time fixture."""
    return datetime.datetime.now() - datetime.timedelta(days=5)


@pytest.fixture(scope='function')
def end_time():
    """A feed stop time fixture."""
    return datetime.datetime.now() + datetime.timedelta(days=1)


@pytest.fixture(scope='function')
def feed_name():
    """A feed name fixture."""
    return 'test'


@pytest.fixture(scope='function')
def feed_download_url():
    """A feed download URL."""
    return 'https://matchlight-reports.s3.amazonaws.com/{}.csv'.format(
        str(uuid.uuid4()))


@pytest.fixture(scope='function')
def feed_payload(feed_name, start_time, end_time):
    """A feed payload fixture."""
    return json.dumps({
        'name': feed_name,
        'description': '',
        'recent_alerts_count': 5,
        'start_timestamp': int(matchlight.utils.datetime_to_unix(start_time)),
        'stop_timestamp': int(matchlight.utils.datetime_to_unix(end_time)),
    })


@pytest.fixture(scope='function')
def feed_report_csv():
    """A CSV-formatted feed report fixture."""
    rows = []
    headers = ('ts', 'artifact_id', 'url', 'value', 'description')
    rows.append(','.join(headers))
    for i in six.moves.range(10):
        rows.append(','.join((
            datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S'),
            uuid.uuid4().hex,
            'http://pastebin.com/5ZLLE3jY',
            'Potato',
            '',
        )))
    return rows


@pytest.fixture(scope='function')
def feed(feed_payload):
    """A feed object fixture."""
    return matchlight.Feed(**json.loads(feed_payload))


def test_permanent_feed(feed_payload):
    """Verifies the end property of a permanent (non expiring) feed."""
    payload = json.loads(feed_payload)

    payload['stop_timestamp'] = None
    assert matchlight.Feed(**payload).end is None

    del payload['stop_timestamp']
    assert matchlight.Feed(**payload).end is None


def test_feed_details(feed):
    """Verifies the feed details mapping."""
    expected = {
        'name': feed.name,
        'description': feed.description,
        'recent_alerts_count': feed.recent_alerts_count,
        'start_timestamp': feed.start_timestamp,
        'stop_timestamp': feed.stop_timestamp,
    }
    assert expected == feed.details


def test_feed_timestamps(feed):
    """Verifies that feed timestamps are parsed correctly."""
    assert datetime.datetime.fromtimestamp(feed.start_timestamp) == feed.start
    assert datetime.datetime.fromtimestamp(feed.stop_timestamp) == feed.end


@responses.activate
def test_feed_counts(connection, feed, start_time, end_time):
    """Verifies requesting feed counts."""
    expected_response = {
        int(time.time()): random.randint(0, 1000),
        int(time.time()): random.randint(0, 1000),
        int(time.time()): random.randint(0, 1000),
        int(time.time()): random.randint(0, 1000),
    }
    responses.add(
        method=responses.POST,
        url='{}/feeds/{}'.format(matchlight.MATCHLIGHT_API_URL_V2, feed.name),
        json=expected_response,
        status=200
    )
    result = connection.feeds.counts(feed, start_time, end_time)
    assert len(responses.calls) == 1
    assert result == connection.feeds._format_count(expected_response)

    result = connection.feeds.counts(feed.name, start_time, end_time)
    assert len(responses.calls) == 2
    assert result == connection.feeds._format_count(expected_response)


@responses.activate
def test_feed_download(connection, feed, start_time, end_time,
                       feed_download_url, feed_report_csv):
    """Verifies feed downloads."""
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/prepare'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'feed_response_id': 1},
        status=200
    )
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/link'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'status': 'pending'},
        status=200
    )
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/link'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'status': 'ready', 'url': feed_download_url},
        status=200
    )
    responses.add(
        method=responses.GET,
        url=feed_download_url,
        content_type='text/csv',
        body='\n'.join(feed_report_csv),
        status=200
    )

    rows = connection.feeds.download(feed, start_time, end_time)
    assert len(responses.calls) == 4
    feed_rows = [
        connection.feeds._format_feed(row)
        for row in csv.DictReader(feed_report_csv)
    ]
    assert rows == feed_rows

    rows = connection.feeds.download(feed.name, start_time, end_time)
    assert len(responses.calls) == 7
    feed_rows = [
        connection.feeds._format_feed(row)
        for row in csv.DictReader(feed_report_csv)
    ]
    assert rows == feed_rows


@responses.activate
def test_feed_download_failed(connection, feed, start_time, end_time):
    """Verifies that feed download failures throw an SDK exception."""
    prepare_url = '{}/feed/{}/prepare'.format(
        matchlight.MATCHLIGHT_API_URL_V2,
        feed.name
    )
    responses.add(
        method=responses.POST,
        url=prepare_url,
        json={'feed_response_id': 1},
        status=200
    )

    link_url = '{}/feed/{}/link'.format(
        matchlight.MATCHLIGHT_API_URL_V2,
        feed.name
    )
    responses.add(
        method=responses.POST,
        url=link_url,
        json={'status': 'pending'},
        status=200
    )
    responses.add(
        method=responses.POST,
        url=link_url,
        json={'status': 'failed', 'message': ''},
        status=200
    )

    with pytest.raises(matchlight.error.SDKError):
        connection.feeds.download(feed, start_time, end_time)
    assert len(responses.calls) == 3

    responses.add(
        method=responses.POST,
        url=prepare_url,
        json='{}',
        status=400
    )
    with pytest.raises(matchlight.error.SDKError):
        connection.feeds.download(feed, start_time, end_time)
    assert len(responses.calls) == 5

    responses.add(
        method=responses.POST,
        url=link_url,
        json={'status': 'potato'},
        status=200
    ),
    with pytest.raises(matchlight.error.SDKError):
        connection.feeds.download(feed, start_time, end_time)
    assert len(responses.calls) == 6


@responses.activate
@mock.patch('io.open', create=True)
def test_feed_download_output(mock_open, connection, feed, start_time,
                              end_time, feed_download_url, feed_report_csv):
    """Verifies feed download writing to a file."""
    mock_open.return_value = mock.MagicMock(spec=io.IOBase)
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/prepare'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'feed_response_id': 1},
        status=200
    )
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/link'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'status': 'pending'},
        status=200
    )
    responses.add(
        method=responses.POST,
        url='{}/feed/{}/link'.format(
            matchlight.MATCHLIGHT_API_URL_V2,
            feed.name
        ),
        json={'status': 'ready', 'url': feed_download_url},
        status=200
    )

    body = '\n'.join(feed_report_csv).encode('utf-8')
    responses.add(
        method=responses.GET,
        url=feed_download_url,
        content_type='text/csv',
        body=body,
        status=200
    )

    connection.feeds.download(
        feed, start_time, end_time, save_path='/tmp/output')
    assert len(responses.calls) == 4
    file_handle = mock_open.return_value.__enter__.return_value
    file_handle.write.assert_called_once_with(body)


@responses.activate
def test_feed_iteration(connection, feed):
    """Verifies feed iteration."""
    responses.add(
        method=responses.GET,
        url='{}/feeds'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'feeds': [feed.details]},
        status=200,
    )
    feeds_iterable = iter(connection.feeds)
    assert next(feeds_iterable).details == feed.details
    with pytest.raises(StopIteration):
        next(feeds_iterable)
    assert len(responses.calls) == 1


@responses.activate
def test_feed_list(connection, feed):
    """Verifies feed listing."""
    responses.add(
        method=responses.GET,
        url='{}/feeds'.format(matchlight.MATCHLIGHT_API_URL_V2),
        json={'feeds': [feed.details]},
        status=200,
    )
    feeds = connection.feeds.all()
    assert len(feeds) == 1
    assert feeds[0].details == feed.details
    assert len(responses.calls) == 1
